from . import Evaluator
from gentab.utils import console, SPINNER, REFRESH, DEVICE

from types import ModuleType
from typing import Type
from typing_extensions import List, Union, Callable, Self

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim
from torch.optim import AdamW
from torch.nn import CrossEntropyLoss
from torch import Tensor
from sklearn.metrics import accuracy_score
from skorch.classifier import NeuralNetClassifier
from skorch.callbacks import EarlyStopping, EpochScoring
from skorch.helper import predefined_split


def reglu(x: Tensor) -> Tensor:
    """The ReGLU activation function from [1].
    References:
        [1] Noam Shazeer, "GLU Variants Improve Transformer", 2020
    """
    assert x.shape[-1] % 2 == 0
    a, b = x.chunk(2, dim=-1)
    return a * F.relu(b)


def geglu(x: Tensor) -> Tensor:
    """The GEGLU activation function from [1].
    References:
        [1] Noam Shazeer, "GLU Variants Improve Transformer", 2020
    """
    assert x.shape[-1] % 2 == 0
    a, b = x.chunk(2, dim=-1)
    return a * F.gelu(b)


class ReGLU(nn.Module):
    """The ReGLU activation function from [shazeer2020glu].

    Examples:
        .. testcode::

            module = ReGLU()
            x = torch.randn(3, 4)
            assert module(x).shape == (3, 2)

    References:
        * [shazeer2020glu] Noam Shazeer, "GLU Variants Improve Transformer", 2020
    """

    def forward(self, x: Tensor) -> Tensor:
        return reglu(x)


class GEGLU(nn.Module):
    """The GEGLU activation function from [shazeer2020glu].

    Examples:
        .. testcode::

            module = GEGLU()
            x = torch.randn(3, 4)
            assert module(x).shape == (3, 2)

    References:
        * [shazeer2020glu] Noam Shazeer, "GLU Variants Improve Transformer", 2020
    """

    def forward(self, x: Tensor) -> Tensor:
        return geglu(x)


def _make_nn_module(module_type: ModuleType, *args) -> nn.Module:
    return (
        (
            ReGLU()
            if module_type == "ReGLU"
            else GEGLU() if module_type == "GEGLU" else getattr(nn, module_type)(*args)
        )
        if isinstance(module_type, str)
        else module_type(*args)
    )


class MLPGorish(nn.Module):
    """The MLP model used in [gorishniy2021revisiting].

    The following scheme describes the architecture:

    .. code-block:: text

          MLP: (in) -> Block -> ... -> Block -> Linear -> (out)
        Block: (in) -> Linear -> Activation -> Dropout -> (out)

    Examples:
        .. testcode::

            x = torch.randn(4, 2)
            module = MLPGorish.make_baseline(x.shape[1], [3, 5], 0.1, 1)
            assert module(x).shape == (len(x), 1)

    References:
        * [gorishniy2021revisiting] Yury Gorishniy, Ivan Rubachev, Valentin Khrulkov, Artem Babenko, "Revisiting Deep Learning Models for Tabular Data", 2021
    """

    class Block(nn.Module):
        """The main building block of `MLP`."""

        def __init__(
            self,
            *,
            d_in: int,
            d_out: int,
            bias: bool,
            activation: ModuleType,
            dropout: float,
        ) -> None:
            super().__init__()
            self.linear = nn.Linear(d_in, d_out, bias)
            self.activation = _make_nn_module(activation)
            self.dropout = nn.Dropout(dropout)

        def forward(self, x: Tensor) -> Tensor:
            return self.dropout(self.activation(self.linear(x)))

    def __init__(
        self,
        *,
        d_in: int,
        d_layers: List[int],
        dropouts: Union[float, List[float]],
        activation: Union[str, Callable[[], nn.Module]],
        d_out: int,
    ) -> None:
        """
        Note:
            `make_baseline` is the recommended constructor.
        """
        super().__init__()
        if isinstance(dropouts, float):
            dropouts = [dropouts] * len(d_layers)
        assert len(d_layers) == len(dropouts)
        assert activation not in ["ReGLU", "GEGLU"]

        self.blocks = nn.ModuleList(
            [
                MLPGorish.Block(
                    d_in=d_layers[i - 1] if i else d_in,
                    d_out=d,
                    bias=True,
                    activation=activation,
                    dropout=dropout,
                )
                for i, (d, dropout) in enumerate(zip(d_layers, dropouts))
            ]
        )
        self.head = nn.Linear(d_layers[-1] if d_layers else d_in, d_out)

    @classmethod
    def make_baseline(
        cls: Type["MLP"],
        d_in: int,
        d_layers: List[int],
        dropout: float,
        d_out: int,
    ) -> "MLPGorish":
        """Create a "baseline" `MLP`.

        This variation of MLP was used in [gorishniy2021revisiting]. Features:

        * :code:`Activation` = :code:`ReLU`
        * all linear layers except for the first one and the last one are of the same dimension
        * the dropout rate is the same for all dropout layers

        Args:
            d_in: the input size
            d_layers: the dimensions of the linear layers. If there are more than two
                layers, then all of them except for the first and the last ones must
                have the same dimension. Valid examples: :code:`[]`, :code:`[8]`,
                :code:`[8, 16]`, :code:`[2, 2, 2, 2]`, :code:`[1, 2, 2, 4]`. Invalid
                example: :code:`[1, 2, 3, 4]`.
            dropout: the dropout rate for all hidden layers
            d_out: the output size
        Returns:
            MLP

        References:
            * [gorishniy2021revisiting] Yury Gorishniy, Ivan Rubachev, Valentin Khrulkov, Artem Babenko, "Revisiting Deep Learning Models for Tabular Data", 2021
        """
        assert isinstance(dropout, float)
        if len(d_layers) > 2:
            assert len(set(d_layers[1:-1])) == 1, (
                "if d_layers contains more than two elements, then"
                " all elements except for the first and the last ones must be equal."
            )
        return MLPGorish(
            d_in=d_in,
            d_layers=d_layers,  # type: ignore
            dropouts=dropout,
            activation="ReLU",
            d_out=d_out,
        )

    def forward(self, x: Tensor) -> Tensor:
        x = x.float()
        for block in self.blocks:
            x = block(x)
        x = self.head(x)
        return x


class MLPClassifier:
    def __init__(
        self,
        input_features,
        num_classes,
        *args,
        epochs: int = 800,
        layers: List[int] = [128, 128],
        dropout: float = 0.1,
        batch_size: int = 8192,
        seed: int = 42,
        lr: float = 1e-5,
        weight_decay: float = 1e-6,
        device: str = DEVICE,
        **kwargs,
    ) -> None:
        self.input_features = input_features
        self.layers = layers
        self.dropout = dropout
        self.num_classes = num_classes
        self.batch_size = batch_size
        self.lr = lr
        self.weigth_decay = weight_decay
        self.epochs = epochs
        self.device = device
        self.seed = seed
        self.acc = accuracy_score
        self.es = EarlyStopping(monitor="train_loss", patience=16)

    def fit(self, X, y) -> Self:
        self.model = MLPGorish.make_baseline(
            self.input_features, self.layers, self.dropout, self.num_classes
        )
        self.net = NeuralNetClassifier(
            self.model,
            criterion=CrossEntropyLoss,
            optimizer=AdamW,
            lr=self.lr,
            optimizer__weight_decay=self.weigth_decay,
            batch_size=self.batch_size,
            max_epochs=self.epochs,
            train_split=None,
            iterator_train__shuffle=True,
            device=self.device,
            callbacks=[self.es, EpochScoring(self.acc, lower_is_better=False)],
            verbose=0,
        )

        self.net.fit(X=X.to_numpy(), y=y)

        return self

    def predict(self, X):
        predictions = self.net.predict_proba(X.to_numpy())

        # Keep original 2D array and get pred. class
        return predictions.argmax(axis=1, keepdims=True)


class MLP(Evaluator):
    def __init__(
        self,
        generator,
        *args,
        layers: List[int] = [128, 128],
        dropout: float = 0.1,
        epochs: int = 1000,
        batch_size: int = 1024,
        **kwargs,
    ) -> None:
        super().__init__(generator, **kwargs)
        self.model = MLPClassifier(
            self.dataset.num_features(),
            self.dataset.num_classes(),
            *args,
            layers,
            dropout,
            epochs=epochs,
            batch_size=batch_size,
            seed=self.seed,
            **kwargs,
        )

    def preprocess(self, X, y, X_test, y_test):
        X = self.dataset.encode_categories(X)
        X_test = self.dataset.encode_categories(X_test)

        # TODO Add ohe to label encoding function
        y = self.generator.dataset.label_encoder_ohe.transform(y)
        y_test = self.generator.dataset.label_encoder_ohe.transform(y_test)

        return X, y, X_test, y_test

    def postprocess(self, pred):
        return self.dataset.decode_labels(pred)



import functools
import torch
import torch.nn as nn
import numpy as np
import random
import math
import torch.nn as nn
import math
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union, cast
from torch import Tensor

from torch.optim import Adam
from torch.optim.lr_scheduler import OneCycleLR
from scipy import integrate

torch.cuda.empty_cache()

###########################################################################################################################################
ModuleType = Union[str, Callable[..., nn.Module]]


class SiLU(nn.Module):
    def forward(self, x):
        return x * torch.sigmoid(x)


def timestep_embedding(timesteps, dim, max_period=10000):
    """
    Create sinusoidal timestep embeddings.

    :param timesteps: a 1-D Tensor of N indices, one per batch element.
                      These may be fractional.
    :param dim: the dimension of the output.
    :param max_period: controls the minimum frequency of the embeddings.
    :return: an [N x dim] Tensor of positional embeddings.
    """
    half = dim // 2
    freqs = torch.exp(
        -math.log(max_period)
        * torch.arange(start=0, end=half, dtype=torch.float32)
        / half
    ).to(device=timesteps.device)
    args = timesteps[:, None].float() * freqs[None]
    embedding = torch.cat([torch.cos(args), torch.sin(args)], dim=-1)
    if dim % 2:
        embedding = torch.cat([embedding, torch.zeros_like(embedding[:, :1])], dim=-1)
    return embedding


def _is_glu_activation(activation: ModuleType):
    return (
        isinstance(activation, str)
        and activation.endswith("GLU")
        or activation in [ReGLU, GEGLU]
    )


def _all_or_none(values):
    assert all(x is None for x in values) or all(x is not None for x in values)


def reglu(x: Tensor) -> Tensor:
    """The ReGLU activation function from [1].
    References:
        [1] Noam Shazeer, "GLU Variants Improve Transformer", 2020
    """
    assert x.shape[-1] % 2 == 0
    a, b = x.chunk(2, dim=-1)
    return a * F.relu(b)


def geglu(x: Tensor) -> Tensor:
    """The GEGLU activation function from [1].
    References:
        [1] Noam Shazeer, "GLU Variants Improve Transformer", 2020
    """
    assert x.shape[-1] % 2 == 0
    a, b = x.chunk(2, dim=-1)
    return a * F.gelu(b)


class ReGLU(nn.Module):
    """The ReGLU activation function from [shazeer2020glu].

    Examples:
        .. testcode::

            module = ReGLU()
            x = torch.randn(3, 4)
            assert module(x).shape == (3, 2)

    References:
        * [shazeer2020glu] Noam Shazeer, "GLU Variants Improve Transformer", 2020
    """

    def forward(self, x: Tensor) -> Tensor:
        return reglu(x)


class GEGLU(nn.Module):
    """The GEGLU activation function from [shazeer2020glu].

    Examples:
        .. testcode::

            module = GEGLU()
            x = torch.randn(3, 4)
            assert module(x).shape == (3, 2)

    References:
        * [shazeer2020glu] Noam Shazeer, "GLU Variants Improve Transformer", 2020
    """

    def forward(self, x: Tensor) -> Tensor:
        return geglu(x)


def _make_nn_module(module_type: ModuleType, *args) -> nn.Module:
    return (
        (
            ReGLU()
            if module_type == "ReGLU"
            else GEGLU() if module_type == "GEGLU" else getattr(nn, module_type)(*args)
        )
        if isinstance(module_type, str)
        else module_type(*args)
    )


class MLP(nn.Module):
    """The MLP model used in [gorishniy2021revisiting].

    The following scheme describes the architecture:

    .. code-block:: text

          MLP: (in) -> Block -> ... -> Block -> Linear -> (out)
        Block: (in) -> Linear -> Activation -> Dropout -> (out)

    Examples:
        .. testcode::

            x = torch.randn(4, 2)
            module = MLP.make_baseline(x.shape[1], [3, 5], 0.1, 1)
            assert module(x).shape == (len(x), 1)

    References:
        * [gorishniy2021revisiting] Yury Gorishniy, Ivan Rubachev, Valentin Khrulkov, Artem Babenko, "Revisiting Deep Learning Models for Tabular Data", 2021
    """

    class Block(nn.Module):
        """The main building block of `MLP`."""

        def __init__(
            self,
            *,
            d_in: int,
            d_out: int,
            bias: bool,
            activation: ModuleType,
            dropout: float,
        ) -> None:
            super().__init__()
            self.linear = nn.Linear(d_in, d_out, bias)
            self.activation = _make_nn_module(activation)
            self.dropout = nn.Dropout(dropout)

        def forward(self, x: Tensor) -> Tensor:
            return self.dropout(self.activation(self.linear(x)))

    def __init__(
        self,
        *,
        d_in: int,
        d_layers: List[int],
        dropouts: Union[float, List[float]],
        activation: Union[str, Callable[[], nn.Module]],
        d_out: int,
    ) -> None:
        """
        Note:
            `make_baseline` is the recommended constructor.
        """
        super().__init__()
        if isinstance(dropouts, float):
            dropouts = [dropouts] * len(d_layers)
        assert len(d_layers) == len(dropouts)
        assert activation not in ["ReGLU", "GEGLU"]

        self.blocks = nn.ModuleList([
            MLP.Block(
                d_in=d_layers[i - 1] if i else d_in,
                d_out=d,
                bias=True,
                activation=activation,
                dropout=dropout,
            )
            for i, (d, dropout) in enumerate(zip(d_layers, dropouts))
        ])
        self.head = nn.Linear(d_layers[-1] if d_layers else d_in, d_out)

    # CODE MLP implementation for tabular
    @classmethod
    def make_baseline(
        cls: Type["MLP"],
        d_in: int,
        d_layers: List[int],
        dropout: float,
        d_out: int,
    ) -> "MLP":
        """Create a "baseline" `MLP`.

        This variation of MLP was used in [gorishniy2021revisiting]. Features:

        * :code:`Activation` = :code:`ReLU`
        * all linear layers except for the first one and the last one are of the same dimension
        * the dropout rate is the same for all dropout layers

        Args:
            d_in: the input size
            d_layers: the dimensions of the linear layers. If there are more than two
                layers, then all of them except for the first and the last ones must
                have the same dimension. Valid examples: :code:`[]`, :code:`[8]`,
                :code:`[8, 16]`, :code:`[2, 2, 2, 2]`, :code:`[1, 2, 2, 4]`. Invalid
                example: :code:`[1, 2, 3, 4]`.
            dropout: the dropout rate for all hidden layers
            d_out: the output size
        Returns:
            MLP

        References:
            * [gorishniy2021revisiting] Yury Gorishniy, Ivan Rubachev, Valentin Khrulkov, Artem Babenko, "Revisiting Deep Learning Models for Tabular Data", 2021
        """
        assert isinstance(dropout, float)
        if len(d_layers) > 2:
            assert len(set(d_layers[1:-1])) == 1, (
                "if d_layers contains more than two elements, then"
                " all elements except for the first and the last ones must be equal."
            )
        return MLP(
            d_in=d_in,
            d_layers=d_layers,  # type: ignore
            dropouts=dropout,
            activation="ReLU",
            d_out=d_out,
        )

    def forward(self, x: Tensor) -> Tensor:
        x = x.float()
        for block in self.blocks:
            x = block(x)
        x = self.head(x)
        return x


###########################################################################################################################################


# f(x,t)
def drift_coeff(x, t, beta_1, beta_0):
    t = torch.tensor(t)
    beta_t = beta_0 + t * (beta_1 - beta_0)
    drift = -0.5 * beta_t * x
    return drift


# g(t)
def diffusion_coeff(t, beta_1, beta_0):
    t = torch.tensor(t)
    beta_t = beta_0 + t * (beta_1 - beta_0)
    diffusion = torch.sqrt(beta_t)
    return diffusion


def marginal_prob_mean(x, t, beta_1, beta_0):
    # x = x.to(device)
    t = torch.tensor(t)
    log_mean_coeff = -0.25 * t**2 * (beta_1 - beta_0) - 0.5 * t * beta_0
    mean = torch.exp(log_mean_coeff)[:, None] * x
    return mean


def marginal_prob_std(t, beta_1, beta_0):
    t = torch.tensor(t)
    log_mean_coeff = -0.25 * t**2 * (beta_1 - beta_0) - 0.5 * t * beta_0
    std = 1 - torch.exp(2.0 * log_mean_coeff)
    return torch.sqrt(std)


drift_coeff_fn = functools.partial(drift_coeff, beta_1=20, beta_0=0.1)
diffusion_coeff_fn = functools.partial(diffusion_coeff, beta_1=20, beta_0=0.1)
marginal_prob_mean_fn = functools.partial(marginal_prob_mean, beta_1=20, beta_0=0.1)
marginal_prob_std_fn = functools.partial(marginal_prob_std, beta_1=20, beta_0=0.1)


def min_max_scaling(factor, scale=(0, 1)):
    std = (factor - factor.min()) / (factor.max() - factor.min())
    new_min = torch.tensor(scale[0])
    new_max = torch.tensor(scale[1])
    return std * (new_max - new_min) + new_min


def compute_v(ll, alpha, beta):
    v = -torch.ones(ll.shape).to(ll.device)
    v[torch.gt(ll, beta)] = torch.tensor(0.0, device=v.device)
    v[torch.le(ll, alpha)] = torch.tensor(1.0, device=v.device)

    if ll[torch.eq(v, -1)].shape[0] != 0 and ll[torch.eq(v, -1)].shape[0] != 1:
        v[torch.eq(v, -1)] = min_max_scaling(ll[torch.eq(v, -1)], scale=(1, 0)).to(
            v.device
        )
    else:
        v[torch.eq(v, -1)] = torch.tensor(0.5, device=v.device)
    return v


def loss_fn(model, Input_Data, T, device, eps=1e-5):
    N, input_dim = Input_Data.shape
    loss_values = torch.empty(N)

    for row in range(N):
        random_t = torch.rand(T) * (1.0 - eps) + eps

        # Compute Perturbed data from SDE
        mean = marginal_prob_mean_fn(Input_Data[row, :], random_t).to(device)
        std = marginal_prob_std_fn(random_t).to(device)
        z = torch.randn(T, input_dim).to(device)
        perturbed_data = mean + z * std[:, None]

        score = model(perturbed_data, random_t).to(device)
        loss_row = torch.mean(torch.sum((score * std[:, None] + z) ** 2, dim=1))

        loss_values[row] = loss_row
    return loss_values.to(device)


class MLPDiffusion(nn.Module):
    def __init__(self, d_in, rtdl_params, dim_t=128):
        super().__init__()
        self.dim_t = dim_t

        rtdl_params["d_in"] = dim_t
        rtdl_params["d_out"] = d_in

        self.mlp = MLP.make_baseline(**rtdl_params)

        self.proj = nn.Linear(d_in, dim_t)
        self.time_embed = nn.Sequential(
            nn.Linear(dim_t, dim_t), nn.SiLU(), nn.Linear(dim_t, dim_t)
        )

    def forward(self, x, timesteps):
        emb = self.time_embed(timestep_embedding(timesteps, self.dim_t))
        x = self.proj(x) + emb
        return self.mlp(x)


def train_diffusion(
    latent_features,
    T,
    eps,
    sigma,
    lr,
    num_batches_per_epoch,
    maximum_learning_rate,
    weight_decay,
    n_epochs,
    batch_size,
    device,
    progress,
    task,
):
    rtdl_params = {
        "d_in": latent_features.shape[1],
        "d_layers": [256, 256],
        "dropout": 0.0,
        "d_out": latent_features.shape[1],
    }

    ScoreNet = MLPDiffusion(latent_features.shape[1], rtdl_params)
    ScoreNet_Parallel = torch.nn.DataParallel(ScoreNet)
    ScoreNet_Parallel = ScoreNet_Parallel.to(device)

    optimizer = Adam(ScoreNet_Parallel.parameters(), lr=lr, weight_decay=weight_decay)
    lr_scheduler = OneCycleLR(
        optimizer,
        max_lr=maximum_learning_rate,
        steps_per_epoch=num_batches_per_epoch,
        epochs=n_epochs,
    )

    losses = []

    for epoch in range(n_epochs):
        batch_idx = random.choices(
            range(latent_features.shape[0]), k=batch_size
        )  ## Choose random indices
        batch_X = latent_features[batch_idx, :]

        loss_values = loss_fn(ScoreNet_Parallel, batch_X, T, device, eps)
        loss = torch.mean(loss_values)

        optimizer.zero_grad()
        loss.backward()
        optimizer.step()
        lr_scheduler.step()

        # Print the training loss over the epoch.
        losses.append(loss.item())

        # tqdm_epoch.set_description("Average Loss: {:5f}".format(loss.item()))
        progress.update(task, completed=epoch + 1)

    return ScoreNet


def Euler_Maruyama_sampling(model, T, N, P, device):
    time_steps = torch.linspace(1.0, 1e-5, T)
    step_size = time_steps[0] - time_steps[1]

    Gen_data = torch.empty(N, P)

    init_x = torch.randn(N, P)
    X = init_x.to(device)

    # tqdm_epoch = tqdm.trange(T)

    with torch.no_grad():
        # TODO Progress here
        for epoch in range(T):
            time_step = time_steps[epoch].unsqueeze(0).to(device)

            # Predictor step (Euler-Maruyama)
            f = drift_coeff_fn(X, time_step).to(device)
            g = diffusion_coeff_fn(time_step).to(device)
            X = (
                X
                - (f - (g**2) * (model(X, time_step))) * step_size.to(device)
                + torch.sqrt(step_size).to(device) * g * torch.randn_like(X).to(device)
            )
            # tqdm_epoch.set_description("Diffusion Level: {:5f}".format(epoch))

    Gen_data = X.cpu()

    return Gen_data



from gentab.utils import console

import json


class Config:
    def __init__(self, path) -> None:
        with open(path, "r") as jsonfile:
            self._config = json.load(jsonfile)

        console.print("✅ Config {} loaded...".format(path))

    def __getitem__(self, key):
        return self._config[key]

    def __setitem__(self, key, data):
        self._config[key] = data

    def exists(self, key: str) -> bool:
        return True if key in self._config else False



from . import Evaluator
from gentab.utils import console, SPINNER, REFRESH

from xgboost import XGBClassifier


class XGBoost(Evaluator):
    def __init__(
        self,
        generator,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(generator, **kwargs)
        self.model = XGBClassifier(
            *args,
            verbosity=0,
            random_state=self.seed,
            # enable_categorical=True,
            # tree_method="hist",
            **kwargs,
        )

    def preprocess(self, X, y, X_test, y_test):
        X = self.dataset.encode_categories(X)
        X_test = self.dataset.encode_categories(X_test)

        y = self.dataset.encode_labels(y)
        y_test = self.dataset.encode_labels(y_test)

        return X, y, X_test, y_test

    def postprocess(self, pred):
        return self.dataset.decode_labels(pred)



from . import Generator
from .ctabg.pipeline.data_preparation import DataPrep
from .ctabg.synthesizer.ctabgan_synthesizer import CTABGANSynthesizer
from gentab.utils import console, PROG_COLUMNS

import pandas as pd
import warnings

warnings.filterwarnings("ignore")


class CTABGAN(Generator):
    """
    Generative model training class based on the CTABGANSynthesizer model

    Variables Constructor:
    1) random_dim -> size of the noise vector fed to the generator
    2) class_dim -> tuple containing dimensionality of hidden layers for the classifier network
    3) num_channels -> no. of channels for deciding respective hidden layers of discriminator and generator networks
    4) l2scale -> parameter to decide strength of regularization of the network based on constraining l2 norm of weights
    5) batch_size -> no. of records to be processed in each mini-batch of training
    6) epochs -> no. of epochs to train the model

    Variables preprocess and generation:
    * Column names can appear in multiple parameters: e.g. capital-gain, integer_columns & mixed_columns
    1) test_ratio -> parameter to choose ratio of size of test to train data
    2) categorical_columns -> list of column names with a categorical distribution (including label)
    3) log_columns -> list of column names with a skewed exponential distribution
       e.g. amount (credit card transactions), most transactions have small amounts,
       very small amount of transactions with very high amounts, fig1b in paper
    4) mixed_columns -> dictionary of column name and categorical modes used for "mix" of numeric and categorical distribution
       e.g. mortgage, it can be no mortgage (0) or mortgage (any value), Fig1a in paper
    5) integer_columns -> list of numeric column names without floating numbers
    6) problem_type -> dictionary of type of ML problem (classification/regression) and target column name
    """

    def __init__(
        self,
        dataset,
        test_ratio=0.20,
        epochs=300,
        class_dim=(256, 256, 256, 256),
        random_dim=100,
        num_channels=64,
        l2scale=1e-5,
        batch_size=8192,
        max_tries_per_batch=4096,
    ) -> None:
        super().__init__(dataset, batch_size, max_tries_per_batch)
        self.epochs = epochs
        self.raw_df = self.dataset.get_single_df()
        self.test_ratio = test_ratio
        self.class_dim = class_dim
        self.random_dim = random_dim
        self.num_channels = num_channels
        self.l2scale = l2scale
        self.categorical_columns = (
            self.config["categorical_columns"]
            + self.config["binary_columns"]
            + [self.config["y_label"]]
        )
        self.log_columns = self.config[str(self)]["log_columns"]
        self.mixed_columns = self.config[str(self)]["mixed_columns"]
        self.integer_columns = self.config["integer_columns"]
        if self.config["task_type"] in ["multiclass", "binary"]:
            self.problem_type = {"Classification": dataset.config["y_label"]}
        else:
            self.problem_type = {"Regression": dataset.config["y_label"]}

        self.synthesizer = CTABGANSynthesizer(
            class_dim=self.class_dim,
            random_dim=self.random_dim,
            num_channels=self.num_channels,
            l2scale=self.l2scale,
            batch_size=self.batch_size,
            epochs=self.epochs,
        )

    def preprocess(self) -> None:
        self.data_prep = DataPrep(
            self.raw_df,
            self.categorical_columns,
            self.log_columns,
            self.mixed_columns,
            self.integer_columns,
            self.problem_type,
            self.test_ratio,
        )

    def train(self) -> None:
        # Setup progress
        self.p.columns = PROG_COLUMNS
        self.p.update(self.gen_task, total=self.epochs)

        # Train
        self.synthesizer.fit(
            train_data=self.data_prep.df,
            categorical=self.data_prep.column_types["categorical"],
            mixed=self.data_prep.column_types["mixed"],
            type=self.problem_type,
            progress=self.p,
            task=self.gen_task,
        )

    def sample(self) -> pd.DataFrame:
        sample = self.synthesizer.sample(self.batch_size)
        return self.data_prep.inverse_prep(sample)



import os
import warnings
import json
import typing as tp
import logging

import numpy as np
import pandas as pd

from tqdm import tqdm

# from peft import get_peft_config, get_peft_model, PrefixTuningConfig, TaskType, PeftType
import torch
from transformers import (
    AutoTokenizer,
    AutoModelForCausalLM,
    TrainingArguments,
    AutoConfig,
    TrainerCallback,
    logging,
)

from .tabula_dataset import TabulaDataset, TabulaDataCollator
from .tabula_start import TabulaStart, CategoricalStart, ContinuousStart, RandomStart
from .tabula_trainer import TabulaTrainer
from .tabula_utils import (
    _convert_tokens_to_dataframe,
    _array_to_dataframe,
    _get_column_distribution,
    _convert_tokens_to_text,
    _convert_text_to_tabular_data,
)

logging.set_verbosity_error()
os.environ["TRANSFORMERS_VERBOSITY"] = "error"


class MyCallback(TrainerCallback):
    "A callback to track progress on epoch end"

    def __init__(self, progress, task) -> None:
        super().__init__()
        self.progress = progress
        self.task = task
        self.epoch = 1

    def on_epoch_end(self, args, state, control, **kwargs):
        self.progress.update(self.task, completed=self.epoch)
        self.epoch += 1


class TabMP:
    """Tabula Class

    The Tabula class handles the whole generation flow. It is used to fine-tune a large language model for tabular data,
    and to sample synthetic tabular data.

    Attributes:
        llm (str): HuggingFace checkpoint of a pretrained large language model, used a basis of our model
        tokenizer (AutoTokenizer): Tokenizer, automatically downloaded from llm-checkpoint
        model (AutoModelForCausalLM): Large language model, automatically downloaded from llm-checkpoint
        experiment_dir (str): Directory, where the training checkpoints will be saved
        epochs (int): Number of epochs to fine-tune the model
        batch_size (int): Batch size used for fine-tuning
        train_hyperparameters (dict): Additional hyperparameters added to the TrainingArguments used by the
         HuggingFaceLibrary, see here the full list of all possible values
         https://huggingface.co/docs/transformers/main/en/main_classes/trainer#transformers.TrainingArguments
        columns (list): List of all features/columns of the tabular dataset
        num_cols (list): List of all numerical features/columns of the tabular dataset
        conditional_col (str): Name of a feature/column on which the sampling can be conditioned
        conditional_col_dist (dict | list): Distribution of the feature/column specified by condtional_col
    """

    def __init__(
        self,
        llm: str,
        experiment_dir: str = "trainer_tabula",
        epochs: int = 100,
        batch_size: int = 8,
        categorical_columns: list = [],
        **train_kwargs,
    ):
        """Initializes Tabula.

        Args:
            llm: HuggingFace checkpoint of a pretrained large language model, used a basis of our model
            experiment_dir:  Directory, where the training checkpoints will be saved
            epochs: Number of epochs to fine-tune the model
            batch_size: Batch size used for fine-tuning
            train_kwargs: Additional hyperparameters added to the TrainingArguments used by the HuggingFaceLibrary,
             see here the full list of all possible values
             https://huggingface.co/docs/transformers/main/en/main_classes/trainer#transformers.TrainingArguments
        """
        # Load Model and Tokenizer from HuggingFace
        self.llm = llm
        self.tokenizer = AutoTokenizer.from_pretrained(self.llm)
        self.tokenizer.pad_token = self.tokenizer.eos_token
        self.config = AutoConfig.from_pretrained(self.llm)
        self.model = AutoModelForCausalLM.from_config(self.config)

        # Set the training hyperparameters
        self.experiment_dir = experiment_dir
        self.epochs = epochs
        self.batch_size = batch_size
        self.train_hyperparameters = train_kwargs
        self.categorical_columns = categorical_columns

        # Needed for the sampling process
        self.columns = None
        self.num_cols = None
        self.conditional_col = None
        self.conditional_col_dist = None
        self.token_list_length = []

    def fit(
        self,
        data: tp.Union[pd.DataFrame, np.ndarray],
        column_names: tp.Optional[tp.List[str]] = None,
        conditional_col: tp.Optional[str] = None,
        resume_from_checkpoint: tp.Union[bool, str] = False,
        progress=None,
        task=None,
    ) -> TabulaTrainer:
        """Fine-tune Tabula using tabular data.

        Args:
            data: Pandas DataFrame or Numpy Array that contains the tabular data
            column_names: If data is Numpy Array, the feature names have to be defined. If data is Pandas
            DataFrame, the value is ignored
            conditional_col: If given, the distribution of this column is saved and used as a starting
            point for the generation process later. If None, the last column is considered as conditional feature
            resume_from_checkpoint: If True, resumes training from the latest checkpoint in the experiment_dir.
            If path, resumes the training from the given checkpoint (has to be a valid HuggingFace checkpoint!)

        Returns:
            TabulaTrainer used for the fine-tuning process
        """
        for var in data.columns:
            if var == data.columns[0]:
                encoded_term = [
                    "%s %s" % (var, str(data[var][i]).strip()) for i in data.index
                ]
                token_list = self.tokenizer(encoded_term)
                self.token_list_length.append(
                    max(len(l) for l in token_list["input_ids"])
                )
            else:
                encoded_term = ["%s" % (str(data[var][i]).strip()) for i in data.index]
                token_list = self.tokenizer(encoded_term)
                self.token_list_length.append(
                    max(len(l) for l in token_list["input_ids"])
                )

        df = _array_to_dataframe(data, columns=column_names)
        self._update_column_information(df)
        self._update_conditional_information(df, conditional_col)

        # Convert DataFrame into HuggingFace dataset object
        tabula_ds = TabulaDataset.from_pandas(df)
        tabula_ds.set_tokenizer(self.tokenizer)

        data_collator = TabulaDataCollator(self.tokenizer)
        data_collator.set_token_list_length(self.token_list_length)
        # tabula_ds.set_token_list_length(self.token_list_length)

        # Set training hyperparameters
        training_args = TrainingArguments(
            self.experiment_dir,
            num_train_epochs=self.epochs,
            per_device_train_batch_size=self.batch_size,
            save_strategy="no",
            **self.train_hyperparameters,
        )
        tabula_trainer = TabulaTrainer(
            self.model,
            training_args,
            train_dataset=tabula_ds,
            tokenizer=self.tokenizer,
            data_collator=data_collator,
        )
        tabula_trainer.add_callback(MyCallback(progress, task))

        # Start training
        tabula_trainer.train(resume_from_checkpoint=resume_from_checkpoint)
        return tabula_trainer

    def sample(
        self,
        n_samples: int,
        start_col: tp.Optional[str] = "",
        start_col_dist: tp.Optional[tp.Union[dict, list]] = None,
        temperature: float = 0.7,
        k: int = 100,
        max_length: int = 100,
        device: str = "cuda",
        max_tries: int = 1338,
        progress=None,
        task=None,
    ) -> pd.DataFrame:
        """Generate synthetic tabular data samples

        Args:
            n_samples: Number of synthetic samples to generate
            start_col: Feature to use as starting point for the generation process. If not given, the target
             learned during the fitting is used as starting point
            start_col_dist: Feature distribution of the starting feature. Should have the format
             "{F1: p1, F2: p2, ...}" for discrete columns or be a list of possible values for continuous columns.
             If not given, the target distribution learned during the fitting is used as starting point
            temperature: The generation samples each token from the probability distribution given by a softmax
             function. The temperature parameter controls the softmax function. A low temperature makes it sharper
             (0 equals greedy search), a high temperature brings more diversity but also uncertainty into the output.
             See this blog article (https://huggingface.co/blog/how-to-generate) to read more about the generation
             process
            k: Sampling Batch Size. Set as high as possible. Speeds up the generation process significantly
            max_length: Maximal number of tokens to generate - has to be long enough to not cut any information!
            device: Set to "cpu" if the GPU should not be used. You can also specify the concrete GPU

        Returns:
            Pandas DataFrame with n_samples rows of generated data
        """
        tabula_start = self._get_start_sampler(start_col, start_col_dist)

        # Move model to device
        self.model.to(device)

        # Init empty DataFrame for the generated samples
        df_gen = pd.DataFrame(columns=self.columns)

        # Start generation process

        already_generated = 0
        tries = 0
        while n_samples > df_gen.shape[0] and tries < max_tries:
            start_tokens = tabula_start.get_start_tokens(k)
            start_tokens = torch.tensor(start_tokens).to(device)

            # Generate tokens
            tokens = self.model.generate(
                input_ids=start_tokens,
                max_length=max_length,
                do_sample=True,
                temperature=temperature,
                pad_token_id=50256,
            )

            df_gen = _convert_tokens_to_dataframe(
                tokens, self.tokenizer, self.token_list_length, self.columns, df_gen
            )

            # Remove rows with flawed numerical values
            for i_num_cols in self.num_cols:
                df_gen = df_gen[
                    pd.to_numeric(df_gen[i_num_cols], errors="coerce").notnull()
                ]

            df_gen[self.num_cols] = df_gen[self.num_cols].astype(float)

            # Remove rows with missing values
            df_gen = df_gen.drop(df_gen[df_gen.isna().any(axis=1)].index)

            # Update process bar
            already_generated = df_gen.shape[0]

            # Update process bar
            progress.update(task, completed=min(already_generated, n_samples))

            tries += 1

        if tries == max_tries:
            raise RuntimeError(
                "max_tries reached, use a higher number, model did not converge or"
                " max_length is not high enough."
            )

        df_gen = df_gen.reset_index(drop=True)
        return df_gen.head(n_samples)

    def tabula_sample(
        self,
        starting_prompts: tp.Union[str, list[str]],
        temperature: float = 0.7,
        max_length: int = 100,
        device: str = "cuda",
    ) -> pd.DataFrame:
        """Generate synthetic tabular data samples conditioned on a given input.

        Args:
            starting_prompts: String or List of Strings on which the output is conditioned.
             For example, "Sex is female, Age is 26"
            temperature: The generation samples each token from the probability distribution given by a softmax
             function. The temperature parameter controls the softmax function. A low temperature makes it sharper
             (0 equals greedy search), a high temperature brings more diversity but also uncertainty into the output.
             See this blog article (https://huggingface.co/blog/how-to-generate) to read more about the generation
             process.
            max_length: Maximal number of tokens to generate - has to be long enough to not cut any information
            device: Set to "cpu" if the GPU should not be used. You can also specify the concrete GPU.

         Returns:
            Pandas DataFrame with synthetic data generated based on starting_prompts
        """

        self.model.to(device)
        starting_prompts = (
            [starting_prompts]
            if isinstance(starting_prompts, str)
            else starting_prompts
        )
        generated_data = []

        # Generate a sample for each starting point
        for prompt in tqdm(starting_prompts):
            start_token = torch.tensor(self.tokenizer(prompt)["input_ids"]).to(device)
            print("start_token: ", start_token)
            # Generate tokens
            gen = self.model.generate(
                input_ids=torch.unsqueeze(start_token, 0),
                max_length=max_length,
                do_sample=True,
                temperature=temperature,
                pad_token_id=50256,
            )
            print("gen: ", gen)
            generated_data.append(torch.squeeze(gen))

        # Convert Text back to Tabular Data
        decoded_data = _convert_tokens_to_text(generated_data, self.tokenizer)
        print("decoded_data: ", decoded_data)
        df_gen = _convert_text_to_tabular_data(
            decoded_data, pd.DataFrame(columns=self.columns)
        )

        return df_gen

    def save(self, path: str):
        """Save Tabula Model

        Saves the model weights and a configuration file in the given directory.

        Args:
            path: Path where to save the model
        """
        # Make directory
        if os.path.isdir(path):
            warnings.warn(f"Directory {path} already exists and is overwritten now.")
        else:
            os.mkdir(path)

        # Save attributes
        with open(path + "/config.json", "w") as f:
            attributes = self.__dict__.copy()
            attributes.pop("tokenizer")
            attributes.pop("model")

            # NDArray is not JSON serializable and therefore has to be converted into a list.
            if isinstance(attributes["conditional_col_dist"], np.ndarray):
                attributes["conditional_col_dist"] = list(
                    attributes["conditional_col_dist"]
                )

            json.dump(attributes, f)

        # Save model weights
        torch.save(self.model.state_dict(), path + "/model.pt")

    def load_finetuned_model(self, path: str):
        """Load fine-tuned model

        Load the weights of a fine-tuned large language model into the Tabula pipeline

        Args:
            path: Path to the fine-tuned model
        """
        self.model.load_state_dict(torch.load(path))

    @classmethod
    def load_from_dir(cls, path: str):
        """Load Tabula class

        Load trained Tabula model from directory.

        Args:
            path: Directory where Tabula model is saved

        Returns:
            New instance of Tabula loaded from directory
        """
        assert os.path.isdir(path), f"Directory {path} does not exist."

        # Load attributes
        with open(path + "/config.json", "r") as f:
            attributes = json.load(f)

        # Create new be_tabula model instance
        tabula = cls(attributes["llm"])

        # Set all attributes
        for k, v in attributes.items():
            setattr(tabula, k, v)

        # Load model weights
        tabula.model.load_state_dict(torch.load(path + "/model.pt", map_location="cpu"))

        return tabula

    def _update_column_information(self, df: pd.DataFrame):
        # Update the column names (and numerical columns for some sanity checks after sampling)
        self.columns = df.columns.to_list()
        self.num_cols = df.select_dtypes(include=np.number).columns.to_list()

    def _update_conditional_information(
        self, df: pd.DataFrame, conditional_col: tp.Optional[str] = None
    ):
        assert conditional_col is None or isinstance(
            conditional_col, str
        ), f"The column name has to be a string and not {type(conditional_col)}"
        assert conditional_col is None or conditional_col in df.columns, (
            f"The column name {conditional_col} is not in the feature names of the"
            " given dataset"
        )

        # Take the distribution of the conditional column for a starting point in the generation process
        self.conditional_col = conditional_col if conditional_col else df.columns[-1]
        self.conditional_col_dist = _get_column_distribution(df, self.conditional_col)

    def _get_start_sampler(
        self,
        start_col: tp.Optional[str],
        start_col_dist: tp.Optional[tp.Union[tp.Dict, tp.List]],
    ) -> TabulaStart:
        if start_col and start_col_dist is None:
            raise ValueError(
                f"Start column {start_col} was given, but no corresponding"
                " distribution."
            )
        if start_col_dist is not None and not start_col:
            raise ValueError(
                f"Start column distribution {start_col} was given, the column name is"
                " missing."
            )

        assert start_col is None or isinstance(
            start_col, str
        ), f"The column name has to be a string and not {type(start_col)}"
        assert (
            start_col_dist is None
            or isinstance(start_col_dist, dict)
            or isinstance(start_col_dist, list)
        ), (
            "The distribution of the start column on has to be a list or a dict and"
            f" not {type(start_col_dist)}"
        )

        start_col = start_col if start_col else self.conditional_col
        start_col_dist = start_col_dist if start_col_dist else self.conditional_col_dist

        if isinstance(start_col_dist, dict):
            return CategoricalStart(self.tokenizer, start_col, start_col_dist)
        elif isinstance(start_col_dist, list):
            return ContinuousStart(self.tokenizer, start_col, start_col_dist)
        else:
            return RandomStart(self.tokenizer, self.columns)



from . import Generator
from gentab.utils import console

import pandas as pd
from imblearn.over_sampling import RandomOverSampler
from collections import Counter


class ROS(Generator):
    def __init__(self, dataset, sampling_strategy="auto", shrinkage=None) -> None:
        super().__init__(dataset)
        self.shrinkage = shrinkage
        self.sampling_strategy = sampling_strategy

    def preprocess(self) -> None:
        super().preprocess()

    def train(self) -> None:
        super().train()

    def sample(self) -> pd.DataFrame:
        return super().sample()

    def resample(self, n_samples, append) -> None:
        if append:
            for cls, cnt in n_samples.items():
                n_samples[cls] += self.orig_counts[cls]

        ros = RandomOverSampler(
            random_state=self.seed,
            sampling_strategy=n_samples,
            shrinkage=self.shrinkage,
        )
        self.dataset.X_gen, self.dataset.y_gen = ros.fit_resample(
            self.dataset.X, self.dataset.y
        )

    def balance(self) -> None:
        ros = RandomOverSampler(
            random_state=self.seed,
            sampling_strategy=self.sampling_strategy,
            shrinkage=self.shrinkage,
        )
        self.dataset.X_gen, self.dataset.y_gen = ros.fit_resample(
            self.dataset.X, self.dataset.y
        )



from . import Tuner
from gentab.generators import CTGAN
from gentab.evaluators import Evaluator
from gentab.utils import console, SPINNER, REFRESH

import optuna


class CTGANTuner(Tuner):
    def __init__(
        self,
        evaluator: Evaluator,
        trials: int,
        *args,
        min_epochs: int = 300,
        max_epochs: int = 600,
        min_batch: int = 512,
        max_batch: int = 4096,
        max_tries_per_batch: int = 8192,
        timeout: int = None,
        **kwargs,
    ) -> None:
        super().__init__(
            evaluator,
            trials,
            min_epochs=min_epochs,
            max_epochs=max_epochs,
            min_batch=min_batch,
            max_batch=max_batch,
            max_tries_per_batch=max_tries_per_batch,
            timeout=timeout,
        )

    def objective(self, trial: optuna.trial.Trial) -> float:
        epochs = trial.suggest_int("epochs", self.min_epochs, self.max_epochs)
        batch_size = trial.suggest_categorical("batch_size", self.batch_sizes)
        discriminator_dim = (
            trial.suggest_categorical("disc_dim_in", [32, 64, 128, 256, 512]),
            trial.suggest_categorical("disc_dim_out", [32, 64, 128, 256, 512]),
        )
        discriminator_decay = trial.suggest_float(
            "discriminator_decay", 1e-7, 1e-5, log=True
        )
        discriminator_lr = trial.suggest_float("discriminator_lr", 2e-5, 2e-3, log=True)
        discriminator_steps = trial.suggest_categorical(
            "discriminator_steps", [1, 2, 4, 8]
        )
        embedding_dim = trial.suggest_categorical("embedding_dim", [64, 128, 256, 512])
        generator_decay = trial.suggest_float("generator_decay", 1e-7, 1e-5, log=True)
        generator_dim = (
            trial.suggest_categorical("gen_dims_in", [32, 64, 128, 256, 512]),
            trial.suggest_categorical("gen_dims_out", [32, 64, 128, 256, 512]),
        )
        generator_lr = trial.suggest_float("discriminator_lr", 2e-5, 2e-3, log=True)

        log_frequency = trial.suggest_categorical("log_frequency", [True, False])
        pac = trial.suggest_categorical("pac", [2, 4, 8, 16, 32])

        self.generator = CTGAN(
            self.dataset,
            epochs=epochs,
            batch_size=batch_size,
            discriminator_dim=discriminator_dim,
            discriminator_decay=discriminator_decay,
            discriminator_lr=discriminator_lr,
            discriminator_steps=discriminator_steps,
            embedding_dim=embedding_dim,
            generator_decay=generator_decay,
            generator_dim=generator_dim,
            generator_lr=generator_lr,
            log_frequency=log_frequency,
            pac=pac,
            max_tries_per_batch=self.max_tries_per_batch,
        )
        self.generator.generate()

        self.store_data(trial)

        acc, mcc = self.evaluator.evaluate(validation=True)

        return mcc



from gentab.evaluators import CatBoost, LightGBM, XGBoost, MLP
from gentab.generators import (
    ROS,
    SMOTE,
    ADASYN,
    TVAE,
    CTGAN,
    GaussianCopula,
    CopulaGAN,
    CTABGAN,
    CTABGANPlus,
    AutoDiffusion,
    ForestDiffusion,
    Tabula,
    GReaT,
)
from gentab.tuners import (
    SMOTETuner,
    ADASYNTuner,
    TVAETuner,
    CTGANTuner,
    GaussianCopulaTuner,
    CopulaGANTuner,
    CTABGANTuner,
    CTABGANPlusTuner,
    AutoDiffusionTuner,
    ForestDiffusionTuner,
    TabulaTuner,
    GReaTTuner,
)
from gentab.data import Config, Dataset
from gentab.utils import console

from pathlib import Path
import os
import json

import pandas as pd
import numpy as np


def timing_to_disk(timing, folder, dataset, generator):
    # Save generator parameters to JSON
    Path(folder).mkdir(parents=True, exist_ok=True)
    path = os.path.join(
        folder,
        str(dataset).lower()
        + "_"
        + str(generator).lower()
        + "_"
        + "baseline"
        + ".json",
    )
    with open(path, "w") as fp:
        json.dump({"train_time": timing[0], "gen_time": timing[1]}, fp, indent=4)


def preproc_adult(dataset):
    dataset.merge_classes({"<=50K": ["<=50K."], ">50K": [">50K."]})
    # dataset.reduce_mem()

    return dataset


configs = [
    ("configs/adult.json", preproc_adult, "Adult"),
]

gens = [
    (SMOTE, "SMOTE \cite{chawla2002smote}", SMOTETuner),
    (ADASYN, "ADASYN \cite{he2008adasyn}", ADASYNTuner),
    (TVAE, "TVAE \cite{xu2019modeling}", TVAETuner),
    (CTGAN, "CTGAN \cite{xu2019modeling}", CTGANTuner),
    (GaussianCopula, "GaussianCopula \cite{patki2016synthetic}", GaussianCopulaTuner),
    (CopulaGAN, "CopulaGAN \cite{xu2019modeling}", CopulaGANTuner),
    (CTABGAN, "CTAB-GAN \cite{zhao2021ctab}", CTABGANTuner),
    (CTABGANPlus, "CTAB-GAN+ \cite{zhao2022ctab}", CTABGANPlusTuner),
    (AutoDiffusion, "AutoDiffusion \cite{suh2023autodiff}", AutoDiffusionTuner),
    (
        ForestDiffusion,
        "ForestDiffusion \cite{jolicoeur2023generating}",
        ForestDiffusionTuner,
    ),
    (GReaT, "GReaT \cite{borisov2022language}", GReaTTuner),
    (Tabula, "Tabula \cite{zhao2023tabula}", TabulaTuner),
]

evals = [(LightGBM, "LightGBM")]

timing_baseline = pd.DataFrame()
timing_tuned = pd.DataFrame()

for c in configs:
    config = Config(c[0])

    dataset = c[1](Dataset(config))
    console.print(dataset.class_counts(), dataset.row_count())

    for g in gens:
        generator = g[0](dataset)
        for e in evals:
            evaluator = e[0](generator)

            tuner = g[2](evaluator, 0)
            tuned = tuner.get_tuning_info()

            timing_tuned.loc[g[1], "Fit"] = tuned["train_time"]
            timing_tuned.loc[g[1], "Sample"] = tuned["gen_time"]

            baseline = generator.benchmark()
            timing_to_disk(baseline, tuner.folder, dataset, generator)
            timing_baseline.loc[g[1], "Fit"] = baseline[0]
            timing_baseline.loc[g[1], "Sample"] = baseline[1]

round = 2
timing_baseline = timing_baseline.round(2)
timing_tuned = timing_tuned.round(2)

console.print(timing_baseline)
console.print(timing_tuned)

min_bl = timing_baseline.min()
min_tn = timing_tuned.min()

lines = []
for (index, bl), (_, tn) in zip(timing_baseline.iterrows(), timing_tuned.iterrows()):
    if min_bl.loc["Fit"] == bl["Fit"]:
        line = (
            index
            + " & "
            + "\\textbf{{{:.{prec}f}}}".format(bl["Fit"], prec=round)
            + " & "
        )
    elif bl["Fit"] != 0.0:
        line = index + " & " + "{:.{prec}f}".format(bl["Fit"], prec=round) + " & "
    else:
        line = index + " & - & "

    if min_bl.loc["Sample"] == bl["Sample"]:
        line += "\\textbf{{{:.{prec}f}}}".format(bl["Sample"], prec=round) + " & "
    elif bl["Sample"] != 0.0:
        line += "{:.{prec}f}".format(bl["Sample"], prec=round) + " & "
    else:
        line += " & - & "

    if min_tn.loc["Fit"] == tn["Fit"]:
        line += "\\textbf{{{:.{prec}f}}}".format(tn["Fit"], prec=round) + " & "
    elif tn["Fit"] != 0.0:
        line += "{:.{prec}f}".format(tn["Fit"], prec=round) + " & "
    else:
        line += " & - & "

    if min_tn.loc["Sample"] == tn["Sample"]:
        line += "\\textbf{{{:.{prec}f}}}".format(tn["Sample"], prec=round) + " \\\\"
    elif tn["Sample"] != 0.0:
        line += "{:.{prec}f}".format(tn["Sample"], prec=round) + " \\\\"
    else:
        line += " - \\\\"

    lines.append(line)

for line in lines:
    console.print(line)


timing_baseline_hours = (timing_baseline / 60 / 60).round(2)
timing_tuned_hours = (timing_tuned / 60 / 60).round(2)

console.print(timing_baseline_hours)
console.print(timing_tuned_hours)

min_bl_hours = timing_baseline_hours.min()
min_tn_hours = timing_tuned_hours.min()


lines = []
for (index, bl), (_, tn), (_, blh), (_, tnh) in zip(
    timing_baseline.iterrows(),
    timing_tuned.iterrows(),
    timing_baseline_hours.iterrows(),
    timing_tuned_hours.iterrows(),
):
    if min_bl_hours.loc["Fit"] == blh["Fit"]:
        line = (
            index
            + " & "
            + "\\textbf{{{:.{prec}f}}}".format(blh["Fit"], prec=round)
            + " & "
        )
    elif blh["Fit"] != 0.0:
        line = index + " & " + "{:.{prec}f}".format(blh["Fit"], prec=round) + " & "
    else:
        line = index + " & - & "

    if min_bl.loc["Sample"] == bl["Sample"]:
        line += "\\textbf{{{:.{prec}f}}}".format(bl["Sample"], prec=round) + " & "
    elif bl["Sample"] != 0.0:
        line += "{:.{prec}f}".format(bl["Sample"], prec=round) + " & "
    else:
        line += " & - & "

    if min_tn_hours.loc["Fit"] == tnh["Fit"]:
        line += "\\textbf{{{:.{prec}f}}}".format(tnh["Fit"], prec=round) + " & "
    elif tnh["Fit"] != 0.0:
        line += "{:.{prec}f}".format(tnh["Fit"], prec=round) + " & "
    else:
        line += " & - & "

    if min_tn.loc["Sample"] == tn["Sample"]:
        line += "\\textbf{{{:.{prec}f}}}".format(tn["Sample"], prec=round) + " \\\\"
    elif tn["Sample"] != 0.0:
        line += "{:.{prec}f}".format(tn["Sample"], prec=round) + " \\\\"
    else:
        line += " - \\\\"

    lines.append(line)

for line in lines:
    console.print(line)



from gentab.evaluators import KNN, LightGBM, XGBoost, MLP
from gentab.generators import (
    ROS,
    SMOTE,
    ADASYN,
    TVAE,
    CTGAN,
    GaussianCopula,
    CopulaGAN,
    CTABGAN,
    CTABGANPlus,
    AutoDiffusion,
    ForestDiffusion,
    Tabula,
    GReaT,
)
from gentab.data import Config, Dataset
from gentab.utils import console

import pandas as pd
import numpy as np


def preproc_playnet(dataset):
    dataset.reduce_size(
        {
            "left_attack": 0.97,
            "right_attack": 0.97,
            "right_transition": 0.9,
            "left_transition": 0.9,
            "time_out": 0.8,
            "left_penal": 0.5,
            "right_penal": 0.5,
        }
    )
    dataset.merge_classes(
        {
            "attack": ["left_attack", "right_attack"],
            "transition": ["left_transition", "right_transition"],
            "penalty": ["left_penal", "right_penal"],
        }
    )
    dataset.reduce_mem()

    return dataset


def preproc_adult(dataset):
    dataset.merge_classes({"<=50K": ["<=50K."], ">50K": [">50K."]})
    dataset.reduce_mem()

    return dataset


def preproc_car(dataset):
    return dataset


configs = [
    ("configs/car_evaluation_cr.json", preproc_car, "Car Evaluation"),
    ("configs/playnet_cr.json", preproc_playnet, "PlayNet"),
    ("configs/adult_cr.json", preproc_adult, "Adult"),
]

gens = [
    (TVAE, "TVAE \cite{xu2019modeling}"),
    (CTGAN, "CTGAN \cite{xu2019modeling}"),
    (GaussianCopula, "GaussianCopula \cite{patki2016synthetic}"),
    (CopulaGAN, "CopulaGAN \cite{xu2019modeling}"),
    (CTABGAN, "CTAB-GAN \cite{zhao2021ctab}"),
    (CTABGANPlus, "CTAB-GAN+ \cite{zhao2022ctab}"),
    (AutoDiffusion, "AutoDiffusion \cite{suh2023autodiff}"),
    (ForestDiffusion, "ForestDiffusion \cite{jolicoeur2023generating}"),
    (GReaT, "GReaT \cite{borisov2022language}"),
    (Tabula, "Tabula \cite{zhao2023tabula}"),
]

jensen_shannon = pd.DataFrame()
wasserstein = pd.DataFrame()
pearson = pd.DataFrame()
theils = pd.DataFrame()
ratio = pd.DataFrame()

for c in configs:
    config = Config(c[0])

    dataset = c[1](Dataset(config))
    console.print(dataset.class_counts(), dataset.row_count())

    for g in gens:
        generator = g[0](dataset)
        try:
            generator.load_from_disk()

            norm_pearson, norm_theils, norm_ratio = dataset.get_correlations()
            pearson.loc[c[2], g[1]] = norm_pearson
            theils.loc[c[2], g[1]] = norm_theils
            ratio.loc[c[2], g[1]] = norm_ratio

            if len(dataset.get_categories()):
                js_dists = dataset.jensen_shannon_distance()
                jensen_shannon.loc[c[2], g[1]] = js_dists.mean()

            if len(dataset.get_continuous()):
                ws_dists = dataset.wasserstein_distance()
                wasserstein.loc[c[2], g[1]] = ws_dists.mean()

        except FileNotFoundError:
            if len(dataset.get_categories()):
                jensen_shannon.loc[c[2], g[1]] = float("inf")
            if len(dataset.get_continuous()):
                wasserstein.loc[c[2], g[1]] = float("inf")

            pearson.loc[c[2], g[1]] = 0.0
            theils.loc[c[2], g[1]] = 0.0
            ratio.loc[c[2], g[1]] = 0.0

round = 2

console.print("Jensen Shannon: \n", jensen_shannon)
console.print("Wasserstein: \n", wasserstein)

js_ranks = jensen_shannon.rank(ascending=True, axis=1)
ws_ranks = wasserstein.rank(ascending=True, axis=1)
console.print("Jensen Shannon: \n", js_ranks)
console.print("Wasserstein \n", ws_ranks)

js_mean = js_ranks.mean().round(round)
js_min = js_mean.min()
ws_mean = ws_ranks.mean().round(round)
ws_min = ws_mean.min()

console.print(js_mean)
console.print(ws_mean)

console.print("Pearson: \n", pearson)
console.print("Theils: \n", theils)
console.print("Ratio: \n", ratio)

ps_mean = pearson.mean().round(round)
ps_min = ps_mean.min()
tu_mean = theils.mean().round(round)
tu_min = tu_mean.min()
rt_mean = ratio.mean().round(round)
rt_min = rt_mean.min()

console.print(ps_mean)
console.print(tu_mean)
console.print(rt_mean)

lines = []
for (index, js), (_, ws), (_, ps), (_, tu), (_, rt) in zip(
    js_mean.items(), ws_mean.items(), ps_mean.items(), tu_mean.items(), rt_mean.items()
):
    if js_min == js:
        line = index + " & " + "\\textbf{{{:.{prec}f}}}".format(js, prec=round) + " & "
    elif js != float("inf"):
        line = index + " & " + "{:.{prec}f}".format(js, prec=round) + " & "
    else:
        line = index + " & - & "

    if ws_min == ws:
        line += "\\textbf{{{:.{prec}f}}}".format(ws, prec=round) + " & "
    elif ws != float("inf"):
        line += "{:.{prec}f}".format(ws, prec=round) + " & "
    else:
        line += " & - & "

    if ps_min == ps:
        line += "\\textbf{{{:.{prec}f}}}".format(ps, prec=round) + " & "
    elif ps != float("inf"):
        line += "{:.{prec}f}".format(ps, prec=round) + " & "
    else:
        line += " & - & "

    if tu_min == tu:
        line += "\\textbf{{{:.{prec}f}}}".format(tu, prec=round) + " & "
    elif tu != float("inf"):
        line += "{:.{prec}f}".format(tu, prec=round) + " & "
    else:
        line += " & - & "

    if rt_min == rt:
        line += "\\textbf{{{:.{prec}f}}}".format(rt, prec=round) + " \\\\"
    elif rt != float("inf"):
        line += "{:.{prec}f}".format(rt, prec=round) + " \\\\"
    else:
        line += "-  \\\\"

    lines.append(line)

for line in lines:
    console.print(line)



import joblib
from .data import DataFrameParser
from .engine import train, featurize
from .network import AutoEncoder


def load(path_to_model_dump):
    dump = joblib.load(path_to_model_dump)
    model = DAE(**dump["constructor_args"])
    model.parser = dump["parser"]
    model.network = AutoEncoder(**model.network_cfg).to(model.device)
    model.network.load_state_dict(dump["network_state_dict"])
    return model


class DAE(object):
    def __init__(
        self,
        body_network="deepstack",
        body_network_cfg=dict(hidden_size=128),
        swap_noise_probas=0.2,
        cats_handling="onehot",
        cards=[],
        embeded_dims=[],
        device="cpu",
    ):
        super().__init__()
        self.body_network = body_network
        self.body_network_cfg = body_network_cfg
        self.swap_noise_probas = swap_noise_probas
        self.cats_handling = cats_handling
        self.cards = cards
        self.embeded_dims = embeded_dims
        self.device = device
        self.network = None
        self.parser = None

    def _parse_dataframe(self, dataframe):
        self.parser = parser = DataFrameParser().fit(dataframe)
        data = parser.transform(dataframe)
        if not self.cards:
            self.cards = parser.cards
        if not self.embeded_dims:
            self.embeded_dims = parser.embeds
        return data

    @property
    def network_cfg(self):
        return dict(
            datatype_info=self.datatype_info,
            cards=self.cards,
            embeded_dims=self.embeded_dims,
            cats_handling=self.cats_handling,
            body_network=self.body_network,
            body_network_cfg=self.body_network_cfg,
        )

    @property
    def datatype_info(self):
        return self.parser.datatype_info()

    @property
    def is_fitted(self):
        return any(val is None for val in [self.network, self.parser])

    def save(self, path_to_model_dump):
        model_state_dict = dict(
            constructor_args=dict(
                body_network=self.body_network,
                body_network_cfg=self.body_network_cfg,
                swap_noise_probas=self.swap_noise_probas,
                cats_handling=self.cats_handling,
                cards=self.cards,
                embeded_dims=self.embeded_dims,
                device=self.device,
            ),
            parser=self.parser,
            network_state_dict=self.network.state_dict(),
        )
        joblib.dump(model_state_dict, path_to_model_dump)

    def fit(
        self,
        dataframe,
        batch_size=32,
        max_epochs=1024,
        validation_ratio=0.2,
        early_stopping_rounds=50,
        verbose=2,
        **train_kwargs,
    ):
        data = self._parse_dataframe(dataframe)

        self.network = train(
            network_cfg_or_network=self.network_cfg,
            data=data,
            datatype_info=self.datatype_info,
            swap_noise_probas=self.swap_noise_probas,
            batch_size=batch_size,
            validation_ratio=validation_ratio,
            early_stopping_rounds=early_stopping_rounds,
            max_epochs=max_epochs,
            device=self.device,
            verbose=verbose,
            **train_kwargs,
        )

    def transform(self, dataframe, batch_size=32):
        assert self.network is not None, "model not trained yet"
        return featurize(
            self.network,
            self.parser.transform(dataframe),
            self.datatype_info,
            batch_size=batch_size,
            device=self.device,
        )

    # def generate(self, N):
    #     z_samples = torch.randn_like(
    #         torch.ones((N, self.encoder.latent_dim)), device=self.device
    #     )
    #     x_gen = self.decoder(z_samples)
    #     x_gen_ = torch.ones_like(x_gen, device=self.device)
    #     i = 0

    #     for v in range(len(self.num_categories)):
    #         x_gen_[
    #             :, i : (i + self.num_categories[v])
    #         ] = torch.distributions.one_hot_categorical.OneHotCategorical(
    #             logits=x_gen[:, i : (i + self.num_categories[v])]
    #         ).sample()
    #         i = i + self.num_categories[v]

    #     x_gen_[:, -self.num_continuous :] = x_gen[
    #         :, -self.num_continuous :
    #     ] + torch.exp(self.noiser(x_gen[:, -self.num_continuous :])) * torch.randn_like(
    #         x_gen[:, -self.num_continuous :]
    #     )
    #     return x_gen_



import numpy as np
import pandas as pd
from sklearn import preprocessing
from sklearn import model_selection


class DataPrep(object):
    
    """
    Data preparation class for pre-processing input data and post-processing generated data

    Variables:
    1) raw_df -> dataframe containing input data
    2) categorical -> list of categorical columns
    3) log -> list of skewed exponential numerical columns
    4) mixed -> dictionary of "mixed" column names with corresponding categorical modes 
    5) integer -> list of numeric columns without floating numbers
    6) type -> dictionary of problem type (i.e classification/regression) and target column
    7) test_ratio -> ratio of size of test to train dataset

    Methods:
    1) __init__() -> instantiates DataPrep object and handles the pre-processing steps for feeding it to the training algorithm
    2) inverse_prep() -> deals with post-processing of the generated data to have the same format as the original dataset


    """


    def __init__(self, raw_df: pd.DataFrame, categorical: list, log:list, mixed:dict, integer:list, type:dict, test_ratio:float):
        
        self.categorical_columns = categorical
        self.log_columns = log
        self.mixed_columns = mixed
        self.integer_columns = integer
        self.column_types = dict()
        self.column_types["categorical"] = []
        self.column_types["mixed"] = {}
        self.lower_bounds = {}
        self.label_encoder_list = []

        # Spliting the input data to obtain training dataset
        target_col = list(type.values())[0]
        y_real = raw_df[target_col]
        X_real = raw_df.drop(columns=[target_col])
        X_train_real, _, y_train_real, _ = model_selection.train_test_split(X_real ,y_real, test_size=test_ratio, stratify=y_real,random_state=42)        
        X_train_real[target_col]= y_train_real

        # Replacing empty strings with na if any and replace na with empty
        self.df = X_train_real
        self.df = self.df.replace(r' ', np.nan)
        self.df = self.df.fillna('empty')
        
        # Dealing with empty values in numeric columns by replacing it with -9999999 and treating it as categorical mode 
        all_columns= set(self.df.columns)
        irrelevant_missing_columns = set(self.categorical_columns)
        relevant_missing_columns = list(all_columns - irrelevant_missing_columns)
        
        for i in relevant_missing_columns:
            if i in list(self.mixed_columns.keys()):
                if "empty" in list(self.df[i].values):
                    self.df[i] = self.df[i].apply(lambda x: -9999999 if x=="empty" else x )
                    self.mixed_columns[i].append(-9999999)
            else:
                if "empty" in list(self.df[i].values):   
                    self.df[i] = self.df[i].apply(lambda x: -9999999 if x=="empty" else x)
                    self.mixed_columns[i] = [-9999999]
        
        # Dealing with skewed exponential numeric distributions by applying log transformation
        if self.log_columns:
            for log_column in self.log_columns:
                # Value added to apply log to non-positive numeric values
                eps = 1 
                # Missing values indicated with -9999999 are skipped
                lower = np.min(self.df.loc[self.df[log_column]!=-9999999][log_column].values) 
                self.lower_bounds[log_column] = lower
                if lower>0: 
                    self.df[log_column] = self.df[log_column].apply(lambda x: np.log(x) if x!=-9999999 else -9999999)
                elif lower == 0:
                    self.df[log_column] = self.df[log_column].apply(lambda x: np.log(x+eps) if x!=-9999999 else -9999999) 
                else:
                    # Negative values are scaled to become positive to apply log
                    self.df[log_column] = self.df[log_column].apply(lambda x: np.log(x-lower+eps) if x!=-9999999 else -9999999)
        
        # Encoding categorical column using label encoding to assign each category within a column with an integer value
        for column_index, column in enumerate(self.df.columns):
            
            if column in self.categorical_columns:        
                label_encoder = preprocessing.LabelEncoder()
                self.df[column] = self.df[column].astype(str)
                label_encoder.fit(self.df[column])
                current_label_encoder = dict()
                current_label_encoder['column'] = column
                current_label_encoder['label_encoder'] = label_encoder
                transformed_column = label_encoder.transform(self.df[column])
                self.df[column] = transformed_column
                self.label_encoder_list.append(current_label_encoder)
                self.column_types["categorical"].append(column_index)
            
            elif column in self.mixed_columns:
                self.column_types["mixed"][column_index] = self.mixed_columns[column]
        
        super().__init__()
        
    def inverse_prep(self, data, eps=1):
        
        # Converting generated data into a dataframe and assign column names as per original dataset
        df_sample = pd.DataFrame(data,columns=self.df.columns)
        
        # Reversing the label encoding assigned to categorical columns according to the original dataset 
        for i in range(len(self.label_encoder_list)):
            le = self.label_encoder_list[i]["label_encoder"]
            df_sample[self.label_encoder_list[i]["column"]] = df_sample[self.label_encoder_list[i]["column"]].astype(int)
            df_sample[self.label_encoder_list[i]["column"]] = le.inverse_transform(df_sample[self.label_encoder_list[i]["column"]])

        # Reversing log by applying exponential transformation with appropriate scaling for non-positive numeric columns 
        # -9999999 used to denote missing values are similarly ignored
        if self.log_columns:
            for i in df_sample:
                if i in self.log_columns:
                    lower_bound = self.lower_bounds[i]
                    if lower_bound>0:
                        df_sample[i].apply(lambda x: np.exp(x) if x!=-9999999 else -9999999) 
                    elif lower_bound==0:
                        df_sample[i] = df_sample[i].apply(lambda x: np.ceil(np.exp(x)-eps) if ((x!=-9999999) & ((np.exp(x)-eps) < 0)) else (np.exp(x)-eps if x!=-9999999 else -9999999))
                    else: 
                        df_sample[i] = df_sample[i].apply(lambda x: np.exp(x)-eps+lower_bound if x!=-9999999 else -9999999)
        
        # Rounding numeric columns without floating numbers in the original dataset
        if self.integer_columns:
            for column in self.integer_columns:
                df_sample[column]= (np.round(df_sample[column].values))
                df_sample[column] = df_sample[column].astype(int)

        # Converting back -9999999 and "empty" to na
        df_sample.replace(-9999999, np.nan,inplace=True)
        df_sample.replace('empty', np.nan,inplace=True)

        return df_sample



from time import perf_counter

from rich.console import Console
from rich.progress import (
    BarColumn,
    MofNCompleteColumn,
    Progress,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
)
import humanize
import datetime as dt

DEVICE = "cuda"
SPINNER = "aesthetic"
REFRESH = 20
EXPAND = False
TRANSIENT = True
IND_COLUMNS = [  # SpinnerColumn(spinner_name=SPINNER),
    TextColumn("🔄 [progress.description]{task.description}"),
    BarColumn(),
]

PROG_COLUMNS = [
    TextColumn("🔄 [progress.description]{task.description}"),
    TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
    BarColumn(),
    MofNCompleteColumn(),
    TextColumn("•"),
    TimeElapsedColumn(),
    TextColumn("•"),
    TimeRemainingColumn(),
]

console = Console()


class ProgressBar:
    def __init__(self, indeterminate: bool = False) -> None:
        # Define custom progress bar
        if indeterminate:
            self.progress = Progress(
                *IND_COLUMNS,
                transient=TRANSIENT,
                console=console,
                expand=EXPAND,
            )
        else:
            self.progress = Progress(
                *PROG_COLUMNS,
                transient=TRANSIENT,
                console=console,
                expand=EXPAND,
            )


class Timer:
    def __init__(self) -> None:
        self._start = None
        self._stop = None
        self.elapsed_s = None
        self.history = []

    def start(self) -> None:
        self._start = perf_counter()

    def stop(self) -> None:
        self._stop = perf_counter()

    def elapsed(self) -> None:
        self.elapsed_s = self._stop - self._start
        self.history.append(self.elapsed_s)
        console.print(
            "⏳ Elapsed time: {}".format(
                humanize.precisedelta(dt.timedelta(seconds=self.elapsed_s))
            )
        )

    def clear_history(self):
        self.history.clear()



import random
import typing as tp

from datasets import Dataset
from dataclasses import dataclass
from transformers import DataCollatorWithPadding
import torch


class TabulaDataset(Dataset):
    """ Tabula Dataset

    The TabulaDataset overwrites the _getitem function of the HuggingFace Dataset Class to include the permutation step.

    Attributes:
        tokenizer (AutoTokenizer): Tokenizer from HuggingFace
    """
    def set_tokenizer(self, tokenizer):
        """ Set the Tokenizer

        Args:
            tokenizer: Tokenizer from HuggingFace
        """
        self.tokenizer = tokenizer


    def _getitem(self, key: tp.Union[int, slice, str], decoded: bool = True, **kwargs) -> tp.Union[tp.Dict, tp.List]:
        """ Get Item from Tabular Data

        Get one instance of the tabular data, permuted, converted to text and tokenized.
        """
        # If int, what else?
        row = self._data.fast_slice(key, 1)

        idx_range = list(range(row.num_columns))
        # random.shuffle(shuffle_idx)
        data_row_text_list = []
        for i in idx_range:
            if i == 0:
                data_row_text_list.append("%s %s" % (row.column_names[i], str(row.columns[i].to_pylist()[0]).strip()))
            # elif i == row.num_columns -1:
            #     # data_row_text_list.append("%s" % (str(row.columns[i].to_pylist()[0]).strip()))
            else:
                data_row_text_list.append("%s" % (str(row.columns[i].to_pylist()[0]).strip()))
        # shuffled_text = ", ".join(
        #     ["%s is %s" % (row.column_names[i], str(row.columns[i].to_pylist()[0]).strip()) for i in shuffle_idx]
        # )
        # print("data row text list: ", data_row_text_list)
        return self.tokenizer(data_row_text_list)['input_ids']
        
    def __getitems__(self, keys: tp.Union[int, slice, str, list]):
        if isinstance(keys, list):
            return [self._getitem(key) for key in keys]
        else:
            return self._getitem(keys)

@dataclass
class TabulaDataCollator(DataCollatorWithPadding):

    """ Tabula Data Collator

    Overwrites the DataCollatorWithPadding to also pad the labels and not only the input_ids

    """

    def set_token_list_length(self, token_list_length):
        """ Set the Tokenizer

        Args:
            token_list_length: a list, each element represents the longest token sequence size for the corresponding column
        """
        self.token_list_length = token_list_length
    


    def __call__(self, features):

        padding_token = 220
        number_sentences = len(features)
        encoding_text = []
        for i in range(len(self.token_list_length)):
            sublist = [item[i] for item in features] # sublist is a list of all the encoding for one data column
            # padded = zip(*itertools.zip_longest(*sublist, fillvalue=50256)) # padding 50256 to the shorter sublist
            if i == (len(self.token_list_length) - 1): # pad 50256 for the last column
                for e in sublist:
                    e.extend([50256] * (self.token_list_length[i] - len(e)))
            else:
                for e in sublist: # pad 220 for the middle column
                    e.extend([padding_token] * (self.token_list_length[i] - len(e)))
            # padded_list =[list(ele) for ele in list(sublist)]
            encoding_text.append(sublist)

        encoded_text = []
        for i in range(number_sentences):
            sentence = [item[i] for item in encoding_text]
            temp = [item for sublist in sentence for item in sublist] # flatten the list
            encoded_text.append(temp)
            
        batch = {'input_ids': torch.Tensor(encoded_text).long(), 'attention_mask': torch.ones(torch.Tensor(encoded_text).shape).long()}
        batch["labels"] = batch["input_ids"].clone()
     
        return batch



from gentab.utils import ProgressBar, console, SPINNER, REFRESH
from gentab import SEED

from typing import Tuple
from sklearn.metrics import (
    precision_recall_fscore_support,
    accuracy_score,
    matthews_corrcoef,
)


class Evaluator:
    def __init__(self, generator, round=2) -> None:
        self.seed = SEED
        self.generator = generator
        self.dataset = generator.dataset
        self.round = round
        self.accuracy = None
        self.macro = None
        self.weighted = None
        self.callbacks = None

    def __str__(self) -> str:
        return self.__class__.__name__

    def preprocess(self, X, y, X_test, y_test):
        return X, y, X_test, y_test

    def postprocess(self, pred):
        return pred

    def compute_accuracy(self, y_true, y_pred):
        return accuracy_score(y_true, y_pred)

    def compute_mcc(self, y_true, y_pred):
        return matthews_corrcoef(y_true, y_pred)

    def compute_f1_p_r(self, y_true, y_pred, average):
        return precision_recall_fscore_support(y_true, y_pred, average=average)

    def compute_metrics(self, X, y, generator, validation) -> None:
        with ProgressBar(indeterminate=True).progress as p:
            eval_task = p.add_task(
                "Evaluating {} accuracy in {}...".format(generator, self.dataset),
                total=None,
            )

            if validation:
                X, y, X_val, y_val = self.preprocess(
                    X, y, self.dataset.X_val, self.dataset.y_val
                )
                if self.callbacks is not None:
                    predictions = self.postprocess(
                        self.model.fit(X, y, callbacks=self.callbacks).predict(X_val)
                    )
                else:
                    predictions = self.postprocess(self.model.fit(X, y).predict(X_val))

                self.accuracy = self.compute_accuracy(self.dataset.y_val, predictions)
                self.mcc = self.compute_mcc(self.dataset.y_val, predictions)
                self.macro = self.compute_f1_p_r(
                    self.dataset.y_val, predictions, "macro"
                )
                self.weighted = self.compute_f1_p_r(
                    self.dataset.y_val, predictions, "weighted"
                )
            else:
                X, y, X_test, y_test = self.preprocess(
                    X, y, self.dataset.X_test, self.dataset.y_test
                )
                if self.callbacks is not None:
                    predictions = self.postprocess(
                        self.model.fit(X, y, callbacks=self.callbacks).predict(X_test)
                    )
                else:
                    predictions = self.postprocess(self.model.fit(X, y).predict(X_test))

                self.accuracy = self.compute_accuracy(self.dataset.y_test, predictions)
                self.mcc = self.compute_mcc(self.dataset.y_test, predictions)
                self.macro = self.compute_f1_p_r(
                    self.dataset.y_test, predictions, "macro"
                )
                self.weighted = self.compute_f1_p_r(
                    self.dataset.y_test, predictions, "weighted"
                )

            console.print(
                "🎯 {} Accuracy: {}".format(
                    generator, round(self.accuracy * 100, self.round - 1)
                )
            )
            console.print(
                "🎯 {} MCC: {}".format(generator, round(self.mcc, self.round))
            )

        console.print(
            "✅ Evaluation complete with {} for {} in {}...".format(
                self, generator, self.dataset
            )
        )

    def evaluate_baseline(self) -> Tuple[float, float]:
        self.compute_metrics(
            self.dataset.X,
            self.dataset.y,
            "Baseline",
            False,
        )

        return self.accuracy, self.mcc

    def evaluate(self, validation=False) -> Tuple[float, float]:
        self.compute_metrics(
            self.dataset.X_gen,
            self.dataset.y_gen,
            self.generator,
            validation,
        )

        return self.accuracy, self.mcc



from . import Tuner
from gentab.generators import GaussianCopula
from gentab.evaluators import Evaluator
from gentab.utils import console, SPINNER, REFRESH

import optuna


class GaussianCopulaTuner(Tuner):
    def __init__(
        self,
        evaluator: Evaluator,
        trials: int,
        *args,
        max_tries_per_batch: int = 8192,
        timeout: int = None,
        **kwargs,
    ) -> None:
        super().__init__(
            evaluator, trials, max_tries_per_batch=max_tries_per_batch, timeout=timeout
        )

    def objective(self, trial: optuna.trial.Trial) -> float:
        default_distribution = trial.suggest_categorical(
            "default_distribution",
            ["norm", "beta", "truncnorm", "uniform", "gamma", "gaussian_kde"],
        )

        self.generator = GaussianCopula(
            self.dataset,
            default_distribution=default_distribution,
            max_tries_per_batch=self.max_tries_per_batch,
        )
        self.generator.generate()

        self.store_data(trial)

        acc, mcc = self.evaluator.evaluate(validation=True)

        return mcc



from . import Tuner
from gentab.generators import ForestDiffusion
from gentab.evaluators import Evaluator
from gentab.utils import console, SPINNER, REFRESH

import optuna


class ForestDiffusionTuner(Tuner):
    def __init__(
        self,
        evaluator: Evaluator,
        trials: int,
        *args,
        min_batch: int = 512,
        max_batch: int = 16384,
        max_tries_per_batch: int = 8192,
        timeout: int = None,
        **kwargs,
    ) -> None:
        super().__init__(
            evaluator,
            trials,
            min_batch=min_batch,
            max_batch=max_batch,
            max_tries_per_batch=max_tries_per_batch,
            timeout=timeout,
        )

    def objective(self, trial: optuna.trial.Trial) -> float:
        # batch_size = trial.suggest_categorical("batch_size", self.batch_sizes)
        n_t = trial.suggest_int("n_t", 20, 80)
        model = trial.suggest_categorical(
            "model", ["xgboost", "random_forest", "lgbm", "catboost"]
        )
        diffusion_type = trial.suggest_categorical("diffusion_type", ["flow", "vp"])
        max_depth = trial.suggest_int("max_depth", 3, 9)
        n_estimators = trial.suggest_int("n_estimators", 20, 200, step=2)
        eta = trial.suggest_float("eta", 1e-8, 1.0, log=True)
        # tree_method = trial.suggest_categorical(
        #     "tree_method", ["exact", "approx", "hist"]
        # )
        reg_lambda = trial.suggest_float("reg_lambda", 1e-8, 1.0, log=True)
        reg_alpha = trial.suggest_float("reg_alpha", 1e-8, 1.0, log=True)
        subsample = trial.suggest_float("subsample", 0.2, 1.0)
        num_leaves = trial.suggest_int("num_leaves", 2, 256)
        duplicate_K = trial.suggest_int("duplicate_K", 2, 50)

        eps = trial.suggest_float("eps", 1e-4, 1e-2, log=True)
        beta_min = trial.suggest_float("beta_min", 1e-8, 0.5, log=True)
        beta_max = trial.suggest_categorical("beta_max", [2, 4, 8, 16])

        self.generator = ForestDiffusion(
            self.dataset,
            n_t=n_t,
            model=model,
            diffusion_type=diffusion_type,
            max_depth=max_depth,
            n_estimators=n_estimators,
            eta=eta,
            # tree_method=tree_method,
            reg_lambda=reg_lambda,
            reg_alpha=reg_alpha,
            subsample=subsample,
            num_leaves=num_leaves,
            duplicate_K=duplicate_K,
            eps=eps,
            beta_min=beta_min,
            beta_max=beta_max,
            # n_batch=batch_size,
            max_tries_per_batch=self.max_tries_per_batch,
        )
        self.generator.generate()

        self.store_data(trial)

        acc, mcc = self.evaluator.evaluate(validation=True)

        return mcc



from . import Generator
from gentab.utils import console

import pandas as pd
from imblearn.over_sampling import SMOTE as sm
from imblearn.over_sampling import SMOTEN as smn
from collections import Counter


class SMOTE(Generator):
    def __init__(self, dataset, k_neighbors=5, sampling_strategy="auto") -> None:
        super().__init__(dataset)
        self.k_neighbors = k_neighbors
        self.sampling_strategy = sampling_strategy

    def preprocess(self) -> None:
        super().preprocess()

    def train(self) -> None:
        super().train()

    def sample(self) -> pd.DataFrame:
        return super().sample()

    def resample(self, n_samples, append) -> None:
        if append:
            for cls, cnt in n_samples.items():
                n_samples[cls] += self.orig_counts[cls]

        if len(self.dataset.X._get_numeric_data().columns) != len(
            self.dataset.X.columns
        ):
            self.dataset.X_gen, self.dataset.y_gen = smn(
                random_state=self.seed,
                sampling_strategy=n_samples,
                k_neighbors=self.k_neighbors,
            ).fit_resample(self.dataset.X, self.dataset.y)
        else:
            self.dataset.X_gen, self.dataset.y_gen = sm(
                random_state=self.seed,
                sampling_strategy=n_samples,
                k_neighbors=self.k_neighbors,
            ).fit_resample(self.dataset.X, self.dataset.y)

    def balance(self) -> None:
        if len(self.dataset.X._get_numeric_data().columns) != len(
            self.dataset.X.columns
        ):
            self.dataset.X_gen, self.dataset.y_gen = smn(
                random_state=self.seed,
                sampling_strategy=self.sampling_strategy,
                k_neighbors=self.k_neighbors,
            ).fit_resample(self.dataset.X, self.dataset.y)
        else:
            self.dataset.X_gen, self.dataset.y_gen = sm(
                random_state=self.seed,
                sampling_strategy=self.sampling_strategy,
                k_neighbors=self.k_neighbors,
            ).fit_resample(self.dataset.X, self.dataset.y)



import numpy as np
from collections import Counter
from torch.utils.data import Dataset


__all__ = [
    "StandardScaler",
    "LabelEncoder",
    "FreqLabelEncoder",
    "DataFrameParser",
    "SingleDataset",
]


class StandardScaler(object):
    def __init__(self):
        self.loc = None
        self.scale = None

    def fit(self, x):
        self.loc, self.scale = np.mean(x), np.std(x)
        return self

    def transform(self, x):
        standardized = (np.array(x) - self.loc) / self.scale
        imputed = np.nan_to_num(standardized)
        return imputed

    def fit_transform(self, x):
        self.fit(x)
        return self.transform(x)


class LabelEncoder(object):
    def __init__(self):
        self.mapping = dict()

    def __len__(self):
        return len(self.mapping)

    def fit(self, x):
        self.mapping = {v: i for i, v in enumerate(set(x))}
        return self

    def transform(self, x):
        return np.array(list(map(self.mapping.__getitem__, x)))


class FreqLabelEncoder(object):
    """A composition of label encoding and frequency encoding. Not reversible."""

    def __init__(self):
        self.freq_counts = None

    def __len__(self):
        return len(self.lbl_encoder)

    def fit(self, x):
        self.freq_counts = Counter(x)
        self.lbl_encoder = LabelEncoder().fit(self.freq_counts.values())
        return self

    def transform(self, x):
        freq_encoded = np.array(list(map(self.freq_counts.__getitem__, x)))
        lbl_encoded = self.lbl_encoder.transform(freq_encoded)
        return lbl_encoded

    def fit_transform(self, x):
        self.fit(x)
        return self.transform(x)


class DataFrameParser(object):
    """Transform dataframe to numpy array for modeling. Not a reversible process.

    It will reshuffle the columns according to datatype: binary->categorical->numerical

    Encoding:
        + Binary variables will be coded as 0, 1
        + Small categorical variables will be label encoded as integers.
        + Categorical variables with large cardinalities will go through count/frequency encoding before label encoding.
        + Numerical will be standardized.

    NaN handling:
        + Fill with mean for numerical. # NODO: Need handling of NaN in categorical? If present in training data is fine.

    """

    def __init__(self, max_cardinality=128):
        self.max_cardinality = max_cardinality
        self.binary_columns = list()
        self.categorical_columns = dict()  # variable name to mode mapping
        self._cards = list()
        self.numerical_columns = list()
        self.need_freq_encoding = set()

    def fit(self, dataframe):
        self._original_order = dataframe.columns.tolist()
        self._original_column_to_dtype = column_to_dtype = dataframe.dtypes.to_dict()

        # sort through columns in dataframe.
        for column, datatype in column_to_dtype.items():
            if datatype in ["O", "<U32"]:
                cardinality = dataframe[column].nunique(dropna=False)
                if cardinality == 2:
                    self.binary_columns.append(column)
                else:
                    self.categorical_columns[column] = dataframe[column].mode()
                    if cardinality > self.max_cardinality:
                        self.need_freq_encoding.add(column)
            elif (
                np.issubdtype(datatype, np.integer) and dataframe[column].nunique() == 2
            ):
                self.binary_columns.append(column)
            else:
                self.numerical_columns.append(column)

        self._column_order = (
            self.binary_columns
            + list(self.categorical_columns.keys())
            + self.numerical_columns
        )

        # fit encoders
        encoders = dict()
        for column in self.binary_columns:
            encoders[column] = LabelEncoder().fit(dataframe[column].astype(str))

        for column in self.categorical_columns:
            if column in self.need_freq_encoding:
                encoders[column] = FreqLabelEncoder().fit(dataframe[column].astype(str))
            else:
                encoders[column] = LabelEncoder().fit(dataframe[column].astype(str))
            self._cards.append(len(encoders[column]))

        for column in self.numerical_columns:
            encoders[column] = StandardScaler().fit(dataframe[column])

        self._embeds = [int(min(600, 1.6 * card**0.5)) for card in self._cards]
        self.encoders = encoders
        return self

    def transform(self, dataframe):
        df = dataframe[self._column_order].copy()
        for column, encoder in self.encoders.items():
            if column in self.numerical_columns:
                df[column] = encoder.transform(df[column])
            else:
                df[column] = encoder.transform(df[column].astype(str))
        return df.values

    def transform_single(self, x):
        output = []
        for column, encoder in self.encoders.items():
            if column in self.numerical_columns:
                output.append(encoder.transform([x[column]])[0])
            else:
                output.append(encoder.transform([str(x[column])])[0])
        return np.array(output)

    @property
    def n_bins(self):
        return len(self.binary_columns)

    @property
    def n_cats(self):
        return len(self.categorical_columns)

    @property
    def n_nums(self):
        return len(self.numerical_columns)

    @property
    def cards(self):
        return self._cards

    @property
    def embeds(self):
        return self._embeds

    def datatype_info(self):
        return {"n_bins": self.n_bins, "n_cats": self.n_cats, "n_nums": self.n_nums}


class SingleDataset(Dataset):
    def __init__(self, data, datatype_info):
        self.data = data
        self.n_bins = datatype_info["n_bins"]
        self.n_cats = datatype_info["n_cats"]
        self.n_nums = datatype_info["n_nums"]

    def __len__(self):
        return self.data.shape[0]

    def __getitem__(self, index):
        single_data = dict()
        if self.n_bins:
            single_data["bins"] = self.data[index, : self.n_bins].astype("float32")
        if self.n_cats:
            single_data["cats"] = self.data[
                index, self.n_bins : self.n_bins + self.n_cats
            ].astype("float32")
        if self.n_nums:
            single_data["nums"] = self.data[index, -self.n_nums :].astype("float32")
        return single_data



import torch
import numpy as np
from collections import Counter
from torch.utils.data import Dataset
import pandas as pd

__all__ = [
    "StandardScaler",
    "LabelEncoder",
    "FreqLabelEncoder",
    "DataFrameParser",
    "SingleDataset",
]


class StandardScaler(object):
    def __init__(self):
        self.loc = None
        self.scale = None

    def fit(self, x):
        self.loc, self.scale = np.mean(x), np.std(x)
        return self

    def transform(self, x):
        standardized = (np.array(x) - self.loc) / self.scale
        imputed = np.nan_to_num(standardized)
        return imputed

    def fit_transform(self, x):
        self.fit(x)
        return self.transform(x)

    def fit_invert(self, x, encoded_col):
        self.loc, self.scale = np.mean(x), np.std(x)
        return encoded_col * self.scale + self.loc


class LabelEncoder(object):
    def __init__(self):
        self.mapping = dict()

    def __len__(self):
        return len(self.mapping)

    def fit(self, x):
        self.mapping = {v: i for i, v in enumerate(set(x))}
        return self

    def fit_bin_int(self, x):
        self.bin_int_encoder = x
        return self

    def fit_transform(self, x):
        return np.array(list(map(self.mapping.__getitem__, x)))

    def fit_int_transform(self, x):
        return np.array(x)

    def fit_invert(self, x, encoded_col):
        self.mapping = {v: i for i, v in enumerate(set(x))}
        inverse_mapping = {v: k for k, v in self.mapping.items()}
        return np.array(list(inverse_mapping[i] for i in encoded_col))


class FreqLabelEncoder(object):
    """A composition of label encoding and frequency encoding. Not reversible."""

    def __init__(self):
        self.freq_counts = None

    def __len__(self):
        return len(self.lbl_encoder)

    def fit(self, x):
        self.freq_counts = Counter(x)
        self.lbl_encoder = LabelEncoder().fit(self.freq_counts.values())
        return self

    def transform(self, x):
        freq_encoded = np.array(list(map(self.freq_counts.__getitem__, x)))
        lbl_encoded = self.lbl_encoder.transform(freq_encoded)
        return lbl_encoded

    def fit_transform(self, x):
        self.fit(x)
        return self.transform(x)


class DataFrameParser(object):
    """Transform dataframe to numpy array for modeling. Not a reversible process.
    It will reshuffle the columns according to datatype: binary->categorical->numerical
    Encoding:
        + Binary variables will be coded as 0, 1
        + Small categorical variables will be label encoded as integers.
        + Categorical variables with large cardinalities will go through count/frequency encoding before label encoding.
        + Numerical will be standardized.
    NaN handling:
        + Fill with mean for numerical. # NODO: Need handling of NaN in categorical? If present in training data is fine.
    """

    def __init__(self, max_cardinality=128):
        self.max_cardinality = max_cardinality
        self.binary_columns = list()
        self.categorical_columns = list()  # variable name to mode mapping
        self._cards = list()
        self.numerical_columns = list()
        self.need_freq_encoding = set()
        self.need_int_encoding = list()
        self.need_bin_int = list()

    def fit(self, dataframe, threshold):
        self.new_dataframe = dataframe.copy()

        self._original_order = self.new_dataframe.columns.tolist()
        self._original_column_to_dtype = (
            column_to_dtype
        ) = self.new_dataframe.dtypes.to_dict()

        # sort through columns in dataframe.
        for column, datatype in column_to_dtype.items():
            if datatype in ["O", "<U32"]:
                cardinality = self.new_dataframe[column].nunique(dropna=False)
                if cardinality == 2:
                    self.binary_columns.append(column)
                else:
                    self.categorical_columns.append(column)
                    if cardinality > self.max_cardinality:
                        self.need_freq_encoding.add(column)

            elif (
                np.issubdtype(datatype, np.integer)
                and self.new_dataframe[column].nunique() == 2
            ):
                self.binary_columns.append(column)
                self.need_bin_int.append(column)

            elif (
                np.issubdtype(datatype, np.integer)
                and self.new_dataframe[column].nunique() <= 25
                and self.new_dataframe[column].nunique() >= 3
            ):
                self.categorical_columns.append(column)
                self.need_int_encoding.append(column)

            else:
                self.numerical_columns.append(column)
                counts = self.new_dataframe[column].value_counts()
                repeated_entries = counts[
                    counts > threshold * len(self.new_dataframe[column])
                ].index.tolist()

                if len(repeated_entries) == 1:
                    new_column_name = "Binary_" + column
                    self.binary_columns.append("Binary_" + column)
                    self.need_bin_int.append("Binary_" + column)
                    self.new_dataframe[new_column_name] = self.new_dataframe[
                        column
                    ].apply(
                        lambda x: (
                            repeated_entries.index(x) + 1
                            if x in repeated_entries
                            else 0
                        )
                    )
                    self.new_dataframe[new_column_name] = self.new_dataframe[
                        new_column_name
                    ].astype(int)

                if len(repeated_entries) >= 2 and len(repeated_entries) <= 25:
                    new_column_name = "Cate_" + column
                    self.categorical_columns.append("Cate_" + column)
                    self.need_int_encoding.append("Cate_" + column)
                    self.new_dataframe[new_column_name] = self.new_dataframe[
                        column
                    ].apply(
                        lambda x: (
                            repeated_entries.index(x) + 1
                            if x in repeated_entries
                            else 0
                        )
                    )

                    self.new_dataframe[new_column_name] = self.new_dataframe[
                        new_column_name
                    ].astype(int)

        self._column_order = (
            self.binary_columns + self.categorical_columns + self.numerical_columns
        )

        # fit encoders
        encoders = dict()
        for column in self.binary_columns:
            if self.new_dataframe[column].dtype == int:
                encoders[column] = LabelEncoder().fit_bin_int(
                    self.new_dataframe[column].astype(int)
                )
            else:
                encoders[column] = LabelEncoder().fit(
                    self.new_dataframe[column].astype(str)
                )

        for column in self.categorical_columns:
            if column in self.need_freq_encoding:
                encoders[column] = FreqLabelEncoder().fit(
                    self.new_dataframe[column].astype(str)
                )
            elif column in self.need_int_encoding:
                encoders[column] = LabelEncoder().fit(
                    self.new_dataframe[column].astype(int)
                )
            else:
                encoders[column] = LabelEncoder().fit(
                    self.new_dataframe[column].astype(str)
                )
            self._cards.append(len(encoders[column]))

        for column in self.numerical_columns:
            encoders[column] = StandardScaler().fit(dataframe[column])

        self._embeds = [int(min(600, 1.6 * card**0.5)) for card in self._cards]
        self.encoders = encoders
        self.new_df = dataframe

        return self

    def transform(self):
        df = self.new_dataframe[self._column_order].copy()
        for column, encoder in self.encoders.items():
            if column in self.numerical_columns:
                df[column] = encoder.fit_transform(df[column])

            elif column in self.need_int_encoding:
                df[column] = encoder.fit_transform(df[column].astype(int))

            elif column in self.need_bin_int:
                df[column] = encoder.fit_int_transform(df[column].astype(int))

            else:
                df[column] = encoder.fit_transform(df[column].astype(str))

        df.columns = self._column_order

        return df.values

    def invert_fit(self, encoded_table):
        decoded_table = encoded_table[self._column_order].copy()
        for column, encoder in self.encoders.items():
            if column in self.numerical_columns:
                decoded_table[column] = StandardScaler().fit_invert(
                    self.new_dataframe[column], encoded_table[column]
                )

            elif (
                column in self.binary_columns
                and np.issubdtype(self.new_dataframe[column].dtype, np.integer) == True
            ):
                decoded_table[column] = encoded_table[column].astype(int)

            elif column in self.need_int_encoding:
                decoded_table[column] = LabelEncoder().fit_invert(
                    self.new_dataframe[column].astype(int), encoded_table[column]
                )

            else:
                decoded_table[column] = LabelEncoder().fit_invert(
                    self.new_dataframe[column].astype(str), encoded_table[column]
                )

        return decoded_table

    @property
    def n_bins(self):
        return len(self.binary_columns)

    @property
    def n_cats(self):
        return len(self.categorical_columns)

    @property
    def n_nums(self):
        return len(self.numerical_columns)

    @property
    def cards(self):
        return self._cards

    @property
    def embeds(self):
        return self._embeds

    def datatype_info(self):
        return {
            "n_bins": self.n_bins,
            "n_cats": self.n_cats,
            "n_nums": self.n_nums,
            "cards": self._cards,
        }

    def column_name(self):
        return self._column_order


def convert_to_table(org_df, gen_output, threshold):
    import torch.nn.functional as F

    def sigmoid_threshold(logits):
        sigmoid_output = torch.sigmoid(logits)
        threshold_output = torch.where(
            sigmoid_output > 0.5, torch.tensor(1), torch.tensor(0)
        )
        return threshold_output

    def softmax_with_max(predictions):
        # Applying softmax function
        probabilities = F.softmax(predictions, dim=1)

        # Getting the index of the maximum element
        max_indices = torch.argmax(probabilities, dim=1)

        return max_indices

    parser_conv = DataFrameParser().fit(org_df, threshold)
    datatype_info = parser_conv.datatype_info()

    n_bins = datatype_info["n_bins"]
    n_cats = datatype_info["n_cats"]
    n_nums = datatype_info["n_nums"]
    cards = datatype_info["cards"]

    cat_tensor = torch.empty(org_df.shape[0], n_cats)
    bin_tensor = torch.empty(org_df.shape[0], n_bins)
    num_tensor = torch.empty(org_df.shape[0], n_nums)

    if n_nums != 0:
        num_tensor = gen_output["nums"].detach()

    if len(cards) != 0:
        for idx in range(len(cards)):
            cat_tensor[:, idx] = (
                softmax_with_max(gen_output["cats"][idx]).detach().to(torch.int64)
            )

    if n_bins != 0:
        for idx in range(n_bins):
            bin_tensor[:, idx] = (
                sigmoid_threshold(gen_output["bins"][:, idx]).detach().to(torch.int64)
            )

    syn_data = torch.cat((bin_tensor, cat_tensor, num_tensor), dim=1)
    t_np = syn_data.numpy()  # convert to Numpy array
    syn_df = pd.DataFrame(t_np)
    syn_df.columns = parser_conv.column_name()
    syn_df = parser_conv.invert_fit(syn_df)

    # Detect pairs of columns with column names 'Cate_X' and 'X'
    cate_column_pairs = [
        (col.replace("Cate_", ""), col)
        for col in syn_df.columns
        if col.startswith("Cate_")
    ]

    ####################################################################################################################
    # Replace the entries in 'X' columns with the corresponding entries in 'Cate_X' columns
    for col, cat_col in cate_column_pairs:
        counts = org_df[col].value_counts()
        repeated_entries = counts[counts > threshold * len(syn_df[col])].index.tolist()

        # Create a mapping between array labels and real numbers
        label_to_number = {
            label: number for label, number in enumerate(repeated_entries, 1)
        }
        array_replaced = np.array([
            label_to_number.get(label, label) for label in syn_df[cat_col]
        ]).astype(int)
        syn_df[col] = np.where(
            array_replaced != 0, array_replaced, syn_df[col].to_numpy()
        )

    # Detect pairs of columns with column names 'Cate_X' and 'X'
    bin_column_pairs = [
        (col.replace("Binary_", ""), col)
        for col in syn_df.columns
        if col.startswith("Binary_")
    ]

    ####################################################################################################################
    # Replace the entries in 'X' columns with the corresponding entries in 'Cate_X' columns
    for col, bin_col in bin_column_pairs:
        counts = org_df[col].value_counts()
        repeated_entries = counts[counts > threshold * len(syn_df[col])].index.tolist()

        # Create a mapping between array labels and real numbers
        label_to_number = {
            label: number for label, number in enumerate(repeated_entries, 1)
        }
        array_replaced = np.array([
            label_to_number.get(label, label) for label in syn_df[bin_col]
        ]).astype(int)
        syn_df[col] = np.where(
            array_replaced != 0, array_replaced, syn_df[col].to_numpy()
        )

    ## Drop the 'Cate_X' columns
    syn_df = syn_df.drop(columns=[cate_col for col, cate_col in cate_column_pairs])
    syn_df = syn_df.drop(columns=[bin_col for col, bin_col in bin_column_pairs])

    orig_column = org_df.columns
    syn_df = syn_df.reindex(columns=orig_column)

    return syn_df



import functools
import torch
import torch.nn as nn
import numpy as np
import random

from torch.optim import Adam
from torch.optim.lr_scheduler import OneCycleLR
from scipy import integrate

torch.cuda.empty_cache()


# f(x,t)
def drift_coeff(x, t, beta_1, beta_0):
    t = torch.tensor(t)
    beta_t = beta_0 + t * (beta_1 - beta_0)
    drift = -0.5 * beta_t * x
    return drift


# g(t)
def diffusion_coeff(t, beta_1, beta_0):
    t = torch.tensor(t)
    beta_t = beta_0 + t * (beta_1 - beta_0)
    diffusion = torch.sqrt(beta_t)
    return diffusion


def marginal_prob_mean(x, t, beta_1, beta_0):
    # x = x.to(device)
    t = torch.tensor(t)
    log_mean_coeff = -0.25 * t**2 * (beta_1 - beta_0) - 0.5 * t * beta_0
    mean = torch.exp(log_mean_coeff)[:, None] * x
    return mean


def marginal_prob_std(t, beta_1, beta_0):
    t = torch.tensor(t)
    log_mean_coeff = -0.25 * t**2 * (beta_1 - beta_0) - 0.5 * t * beta_0
    std = 1 - torch.exp(2.0 * log_mean_coeff)
    return torch.sqrt(std)


drift_coeff_fn = functools.partial(drift_coeff, beta_1=20, beta_0=0.1)
diffusion_coeff_fn = functools.partial(diffusion_coeff, beta_1=20, beta_0=0.1)
marginal_prob_mean_fn = functools.partial(marginal_prob_mean, beta_1=20, beta_0=0.1)
marginal_prob_std_fn = functools.partial(marginal_prob_std, beta_1=20, beta_0=0.1)


def min_max_scaling(factor, scale=(0, 1)):
    std = (factor - factor.min()) / (factor.max() - factor.min())
    new_min = torch.tensor(scale[0])
    new_max = torch.tensor(scale[1])
    return std * (new_max - new_min) + new_min


def compute_v(ll, alpha, beta):
    v = -torch.ones(ll.shape).to(ll.device)
    v[torch.gt(ll, beta)] = torch.tensor(0.0, device=v.device)
    v[torch.le(ll, alpha)] = torch.tensor(1.0, device=v.device)

    if ll[torch.eq(v, -1)].shape[0] != 0 and ll[torch.eq(v, -1)].shape[0] != 1:
        v[torch.eq(v, -1)] = min_max_scaling(ll[torch.eq(v, -1)], scale=(1, 0)).to(
            v.device
        )
    else:
        v[torch.eq(v, -1)] = torch.tensor(0.5, device=v.device)
    return v


def loss_fn(model, Input_Data, T, eps=1e-5):
    N, input_dim = Input_Data.shape
    loss_values = torch.empty(N)

    for row in range(N):
        random_t = torch.rand(T) * (1.0 - eps) + eps

        # Compute Perturbed data from SDE
        mean = marginal_prob_mean_fn(Input_Data[row, :], random_t).to(device)
        std = marginal_prob_std_fn(random_t).to(device)
        z = torch.randn(T, input_dim).to(device)
        perturbed_data = mean + z * std[:, None]

        score = model(perturbed_data, random_t).to(device)
        loss_row = torch.mean(torch.sum((score * std[:, None] + z) ** 2, dim=1))

        loss_values[row] = loss_row
    return loss_values.to(device)


class ConcatSquash(nn.Module):
    def __init__(self, input_dim, output_dim):
        super().__init__()
        self.linear_h = nn.Linear(input_dim, output_dim)
        self.hyper_bias = nn.Linear(1, output_dim, bias=False)
        self.hyper_gate = nn.Linear(1, output_dim, bias=False)

    def forward(self, x, t):
        return self.linear_h(x) * torch.sigmoid(
            self.hyper_bias(t.view(-1, 1)) + self.hyper_gate(t.view(-1, 1))
        )


class TabNetwork(nn.Module):
    def __init__(self, hidden_dims, converted_table_dim):
        super().__init__()

        self.hidden_dims = hidden_dims
        self.converted_table_dim = converted_table_dim

        modules = []
        dim = self.converted_table_dim

        for item in list(self.hidden_dims):
            modules.append(ConcatSquash(dim, dim + item))
            dim = dim + item
            modules.append(nn.ELU())

        modules.append(nn.Linear(dim, self.converted_table_dim))
        self.all_modules = nn.ModuleList(modules)
        self.marginal_prob_std_fn = marginal_prob_std_fn

    def forward(self, x, t):
        modules = self.all_modules
        temb = x
        time = t
        m_idx = 0

        for _ in range(len(self.hidden_dims)):
            temb1 = modules[m_idx](x=temb, t=time)
            m_idx += 1
            temb = modules[m_idx](temb1)
            m_idx += 1

        h = modules[m_idx](temb)
        score_net = h / self.marginal_prob_std_fn(t)[:, None]

        return score_net


def train_diffusion(
    latent_features,
    T,
    hidden_dims,
    converted_table_dim,
    eps,
    sigma,
    lr,
    num_batches_per_epoch,
    maximum_learning_rate,
    weight_decay,
    n_epochs,
    batch_size,
    device,
):
    ScoreNet = TabNetwork(hidden_dims, converted_table_dim)  # Stasy Architecture
    ScoreNet_Parallel = torch.nn.DataParallel(ScoreNet)
    ScoreNet_Parallel = ScoreNet_Parallel.to(device)

    optimizer = Adam(ScoreNet_Parallel.parameters(), lr=lr, weight_decay=weight_decay)
    lr_scheduler = OneCycleLR(
        optimizer,
        max_lr=maximum_learning_rate,
        steps_per_epoch=num_batches_per_epoch,
        epochs=n_epochs,
    )

    # tqdm_epoch = tqdm.trange(n_epochs)
    losses = []

    # alpha0 = 0.25
    # beta0 = 0.95

    # TODO Progress here
    for epoch in range(n_epochs):
        batch_idx = random.choices(
            range(latent_features.shape[0]), k=batch_size
        )  ## Choose random indices
        batch_X = latent_features[batch_idx, :]

        loss_values = loss_fn(ScoreNet_Parallel, batch_X, T, eps)

        # q_alpha = torch.tensor(alpha0 + torch.log( torch.tensor(1+0.0001718*epoch* (1-alpha0), dtype=torch.float32))).clamp_(max=1).to(device)
        # q_beta = torch.tensor(beta0 + torch.log( torch.tensor(1+0.0001718*epoch* (1-beta0), dtype=torch.float32) )).clamp_(max=1).to(device)

        # alpha = torch.quantile(loss_values, q_alpha)
        # beta = torch.quantile(loss_values, q_beta)
        # assert alpha <= beta
        # v = compute_v(loss_values, alpha, beta)

        # loss = torch.mean(v*loss_values)
        loss = torch.mean(loss_values)

        optimizer.zero_grad()
        loss.backward()
        optimizer.step()
        lr_scheduler.step()

        # Print the training loss over the epoch.
        losses.append(loss.item())
        # tqdm_epoch.set_description("Average Loss: {:5f}".format(loss.item()))

    return ScoreNet


def Euler_Maruyama_sampling(model, T, N, P, device):
    time_steps = torch.linspace(1.0, 1e-5, T)
    step_size = time_steps[0] - time_steps[1]

    Gen_data = torch.empty(N, P)

    init_x = torch.randn(N, P)
    X = init_x.to(device)

    # tqdm_epoch = tqdm.trange(T)

    # TODO Progress here
    with torch.no_grad():
        for epoch in range(T):
            time_step = time_steps[epoch].unsqueeze(0).to(device)

            # Predictor step (Euler-Maruyama)
            f = drift_coeff_fn(X, time_step).to(device)
            g = diffusion_coeff_fn(time_step).to(device)
            X = (
                X
                - (f - (g**2) * (model(X, time_step))) * step_size.to(device)
                + torch.sqrt(step_size).to(device) * g * torch.randn_like(X).to(device)
            )
            # tqdm_epoch.set_description("Diffusion Level: {:5f}".format(epoch))

    Gen_data = X.cpu()

    return Gen_data


###############################################################################################################################################


def prob_flow(score, T, N, P, device):
    init_x = torch.randn(N, P, device=device)

    time = np.linspace(1.0, 1e-5, T)
    shape = init_x.shape

    def score_eval_wrapper(sample, time_steps):
        """A wrapper of the score-based model for use by the ODE solver."""
        shape = sample.shape
        sample = torch.tensor(sample, device=device, dtype=torch.float32).reshape(shape)
        time_steps = torch.tensor(
            time_steps, device=device, dtype=torch.float32
        ).reshape((sample.shape[0],))
        with torch.no_grad():
            score_wrap = score(sample, time_steps)
        return score_wrap.cpu().numpy().reshape((-1,)).astype(np.float64)

    def ode_func(t, x):
        """The ODE function for use by the ODE solver."""
        time_steps = np.ones((shape[0],)) * t

        t = torch.tensor(t).to(device)
        x = torch.tensor(x).to(device)
        f = drift_coeff_fn(x, t).cpu().numpy().reshape((-1,))
        g = diffusion_coeff_fn(t).cpu().numpy()

        x = x.reshape(shape)
        return f - 0.5 * (g**2) * score_eval_wrapper(x, time_steps)

    res = integrate.solve_ivp(
        ode_func,
        (1.0, 1e-5),
        init_x.reshape(-1).cpu().numpy(),
        t_eval=time,
        rtol=1e-5,
        atol=1e-5,
        method="RK45",
    )
    x = torch.tensor(res.y[:, -1], device=device).reshape(shape)

    return x.float().cpu()


###############################################################################################################################################


def to_flattened_numpy(x):
    """Flatten a torch tensor `x` and convert it to numpy."""
    return x.detach().cpu().numpy().reshape((-1,))


def from_flattened_numpy(x, shape):
    """Form a torch tensor with the given `shape` from a flattened numpy array `x`."""
    return torch.from_numpy(x.reshape(shape))


def drift_fn(model, x, vec_t):
    N, P = x.shape

    mean = marginal_prob_mean_fn(x, vec_t).to(device)
    std = marginal_prob_std_fn(vec_t).to(device)
    z = torch.randn(N, P).to(device)
    perturbed_data = mean + z * std[:, None]
    score_eval = model(perturbed_data, vec_t).to(device)

    f = drift_coeff_fn(perturbed_data, vec_t)
    g = diffusion_coeff_fn(vec_t)
    drift_fn = f - 0.5 * (g**2) * score_eval

    return drift_fn


def get_div_fn(fn):
    """Create the divergence function of `fn` using the Hutchinson-Skilling trace estimator."""

    def div_fn(x, t, eps):
        grad_fn_eps_list = []
        for epsilon in eps:
            with torch.enable_grad():
                x.requires_grad_(True)
                fn_eps = torch.sum(fn(x, t) * epsilon)
                grad_fn_eps = torch.autograd.grad(fn_eps, x)[0]

            x.requires_grad_(False)
            grad_fn_eps_list.append(
                torch.sum(grad_fn_eps * epsilon, dim=tuple(range(1, len(x.shape))))
            )

        return torch.mean(torch.stack(grad_fn_eps_list), 0)

    return div_fn


def div_fn(model, x, t, noise):
    return get_div_fn(lambda xx, tt: drift_fn(model, xx, tt))(x, t, noise)


def ode_func(t, x):
    sample = from_flattened_numpy(x[: -shape[0]], shape).to(device).type(torch.float32)
    vec_t = torch.ones(sample.shape[0], device=device) * t
    drift = to_flattened_numpy(drift_fn(score, sample, vec_t))
    logp_grad = to_flattened_numpy(div_fn(score, sample, vec_t, epsilon))
    return np.concatenate([drift, logp_grad], axis=0)


# init = np.concatenate([to_flattened_numpy(data), np.zeros((shape[0],))], axis = 0)
# res = integrate.solve_ivp(ode_func, (1e-5, 1.), init, rtol=1e-5, atol=1e-5, method='RK45')



import os
import warnings
import json
import typing as tp
import logging

import numpy as np
import pandas as pd
from sklearn import preprocessing

from tqdm import tqdm
import torch
from transformers import (
    AutoTokenizer,
    AutoModelForCausalLM,
    TrainingArguments,
    AutoConfig,
    TrainerCallback,
    logging,
)

from .tabula_dataset import TabulaDataset, TabulaDataCollator
from .tabula_start import (
    TabulaStart,
    CategoricalStart,
    ContinuousStart,
    RandomStart,
)
from .tabula_trainer import TabulaTrainer
from .tabula_utils import (
    _array_to_dataframe,
    _get_column_distribution,
    _convert_tokens_to_text,
    _convert_text_to_tabular_data,
)

logging.set_verbosity_error()
os.environ["TRANSFORMERS_VERBOSITY"] = "error"


class MyCallback(TrainerCallback):
    "A callback to track progress on epoch end"

    def __init__(self, progress, task) -> None:
        super().__init__()
        self.progress = progress
        self.task = task
        self.epoch = 1

    def on_epoch_end(self, args, state, control, **kwargs):
        self.progress.update(self.task, completed=self.epoch)
        self.epoch += 1


class Tab:
    """Tabula Class

    The Tabula class handles the whole generation flow. It is used to fine-tune a large language model for tabular data,
    and to sample synthetic tabular data.

    Attributes:
        llm (str): HuggingFace checkpoint of a pretrained large language model, used a basis of our model
        tokenizer (AutoTokenizer): Tokenizer, automatically downloaded from llm-checkpoint
        model (AutoModelForCausalLM): Large language model, automatically downloaded from llm-checkpoint
        experiment_dir (str): Directory, where the training checkpoints will be saved
        epochs (int): Number of epochs to fine-tune the model
        batch_size (int): Batch size used for fine-tuning
        train_hyperparameters (dict): Additional hyperparameters added to the TrainingArguments used by the
         HuggingFaceLibrary, see here the full list of all possible values
         https://huggingface.co/docs/transformers/main/en/main_classes/trainer#transformers.TrainingArguments
        columns (list): List of all features/columns of the tabular dataset
        num_cols (list): List of all numerical features/columns of the tabular dataset
        conditional_col (str): Name of a feature/column on which the sampling can be conditioned
        conditional_col_dist (dict | list): Distribution of the feature/column specified by condtional_col
    """

    def __init__(
        self,
        llm: str,
        experiment_dir: str = "trainer_tabula",
        epochs: int = 100,
        batch_size: int = 8,
        categorical_columns: list = [],
        **train_kwargs,
    ):
        """Initializes Tabula.

        Args:
            llm: HuggingFace checkpoint of a pretrained large language model, used a basis of our model
            experiment_dir:  Directory, where the training checkpoints will be saved
            epochs: Number of epochs to fine-tune the model
            batch_size: Batch size used for fine-tuning
            train_kwargs: Additional hyperparameters added to the TrainingArguments used by the HuggingFaceLibrary,
             see here the full list of all possible values
             https://huggingface.co/docs/transformers/main/en/main_classes/trainer#transformers.TrainingArguments
        """
        # Load Model and Tokenizer from HuggingFace
        self.llm = llm
        self.tokenizer = AutoTokenizer.from_pretrained(self.llm)
        self.tokenizer.pad_token = self.tokenizer.eos_token
        self.config = AutoConfig.from_pretrained(self.llm)
        self.model = AutoModelForCausalLM.from_config(self.config)

        # Set the training hyperparameters
        self.experiment_dir = experiment_dir
        self.epochs = epochs
        self.batch_size = batch_size
        self.categorical_columns = categorical_columns
        self.train_hyperparameters = train_kwargs

        # Needed for the sampling process
        self.columns = None
        self.num_cols = None
        self.conditional_col = None
        self.conditional_col_dist = None

    def encode_categorical_column(self, data: pd.DataFrame):
        self.label_encoder_list = []
        for column_index, column in enumerate(data.columns):
            if column in self.categorical_columns:
                label_encoder = preprocessing.LabelEncoder()
                data[column] = data[column].astype(str)
                label_encoder.fit(data[column])
                current_label_encoder = dict()
                current_label_encoder["column"] = column
                current_label_encoder["label_encoder"] = label_encoder
                transformed_column = label_encoder.transform(data[column])
                data[column] = transformed_column
                self.label_encoder_list.append(current_label_encoder)
        return data

    def decode_categorical_column(self, data: pd.DataFrame):

        for i in range(len(self.label_encoder_list)):
            le = self.label_encoder_list[i]["label_encoder"]
            allowed_values = list(range(len(le.classes_)))

            # delete rows that should generate numeric value but generate other data type
            data[self.label_encoder_list[i]["column"]] = pd.to_numeric(
                data[self.label_encoder_list[i]["column"]], errors="coerce"
            )
            data = data.dropna(subset=[self.label_encoder_list[i]["column"]])

            # delete rows that generate category that is out of boundary
            data[self.label_encoder_list[i]["column"]] = data[
                self.label_encoder_list[i]["column"]
            ].astype(float)
            data = data[data[self.label_encoder_list[i]["column"]].isin(allowed_values)]

        for i in range(len(self.label_encoder_list)):
            le = self.label_encoder_list[i]["label_encoder"]
            data[self.label_encoder_list[i]["column"]] = data[
                self.label_encoder_list[i]["column"]
            ].astype(int)
            data[self.label_encoder_list[i]["column"]] = le.inverse_transform(
                data[self.label_encoder_list[i]["column"]]
            )

        return data

    def init(
        self,
        data: tp.Union[pd.DataFrame, np.ndarray],
        column_names: tp.Optional[tp.List[str]] = None,
        conditional_col: tp.Optional[str] = None,
    ):
        df = _array_to_dataframe(data, columns=column_names)
        self._update_column_information(df)
        self._update_conditional_information(df, conditional_col)

        # encoding text to ID for categorical column
        if self.categorical_columns != []:
            df = self.encode_categorical_column(df)

        # Convert DataFrame into HuggingFace dataset object
        # logging.info("Convert data into HuggingFace dataset object...")
        tabula_ds = TabulaDataset.from_pandas(df)
        tabula_ds.set_tokenizer(self.tokenizer)

        # Set training hyperparameters
        # logging.info("Create Tabula Trainer...")
        training_args = TrainingArguments(
            self.experiment_dir,
            num_train_epochs=self.epochs,
            per_device_train_batch_size=self.batch_size,
            save_strategy="no",
            **self.train_hyperparameters,
        )
        tabula_trainer = TabulaTrainer(
            self.model,
            training_args,
            train_dataset=tabula_ds,
            tokenizer=self.tokenizer,
            data_collator=TabulaDataCollator(self.tokenizer),
        )

        return tabula_trainer

    def fit(
        self,
        data: tp.Union[pd.DataFrame, np.ndarray],
        column_names: tp.Optional[tp.List[str]] = None,
        conditional_col: tp.Optional[str] = None,
        resume_from_checkpoint: tp.Union[bool, str] = False,
        progress=None,
        task=None,
    ) -> TabulaTrainer:
        """Fine-tune Tabula using tabular data.

        Args:
            data: Pandas DataFrame or Numpy Array that contains the tabular data
            column_names: If data is Numpy Array, the feature names have to be defined. If data is Pandas
            DataFrame, the value is ignored
            conditional_col: If given, the distribution of this column is saved and used as a starting
            point for the generation process later. If None, the last column is considered as conditional feature
            resume_from_checkpoint: If True, resumes training from the latest checkpoint in the experiment_dir.
            If path, resumes the training from the given checkpoint (has to be a valid HuggingFace checkpoint!)

        Returns:
            TabulaTrainer used for the fine-tuning process
        """
        tabula_trainer = self.init(
            data, column_names=column_names, conditional_col=conditional_col
        )
        tabula_trainer.add_callback(MyCallback(progress, task))
        # Start training
        # logging.info("Start training...")
        tabula_trainer.train(resume_from_checkpoint=resume_from_checkpoint)
        return tabula_trainer

    def sample(
        self,
        n_samples: int,
        start_col: tp.Optional[str] = "",
        start_col_dist: tp.Optional[tp.Union[dict, list]] = None,
        temperature: float = 0.7,
        k: int = 100,
        max_length: int = 100,
        device: str = "cuda",
        max_tries: int = 1338,
        progress=None,
        task=None,
    ) -> pd.DataFrame:
        """Generate synthetic tabular data samples

        Args:
            n_samples: Number of synthetic samples to generate
            start_col: Feature to use as starting point for the generation process. If not given, the target
             learned during the fitting is used as starting point
            start_col_dist: Feature distribution of the starting feature. Should have the format
             "{F1: p1, F2: p2, ...}" for discrete columns or be a list of possible values for continuous columns.
             If not given, the target distribution learned during the fitting is used as starting point
            temperature: The generation samples each token from the probability distribution given by a softmax
             function. The temperature parameter controls the softmax function. A low temperature makes it sharper
             (0 equals greedy search), a high temperature brings more diversity but also uncertainty into the output.
             See this blog article (https://huggingface.co/blog/how-to-generate) to read more about the generation
             process
            k: Sampling Batch Size. Set as high as possible. Speeds up the generation process significantly
            max_length: Maximal number of tokens to generate - has to be long enough to not cut any information!
            device: Set to "cpu" if the GPU should not be used. You can also specify the concrete GPU

        Returns:
            Pandas DataFrame with n_samples rows of generated data
        """
        tabula_start = self._get_start_sampler(start_col, start_col_dist)

        # Move model to device
        self.model.to(device)

        # Init empty DataFrame for the generated samples
        df_gen = pd.DataFrame(columns=self.columns)

        # Start generation process
        already_generated = 0
        tries = 0
        while n_samples > df_gen.shape[0] and tries < max_tries:
            start_tokens = tabula_start.get_start_tokens(k)
            start_tokens = torch.tensor(start_tokens).to(device)

            # Generate tokens
            tokens = self.model.generate(
                input_ids=start_tokens,
                max_length=max_length,
                do_sample=True,
                temperature=temperature,
                pad_token_id=50256,
            )

            # Convert tokens back to tabular data
            text_data = _convert_tokens_to_text(tokens, self.tokenizer)

            df_gen = _convert_text_to_tabular_data(text_data, df_gen)

            # Remove rows with flawed numerical values
            for i_num_cols in self.num_cols:
                df_gen = df_gen[
                    pd.to_numeric(df_gen[i_num_cols], errors="coerce").notnull()
                ]

            df_gen[self.num_cols] = df_gen[self.num_cols].astype(float)

            # Remove rows with missing values
            df_gen = df_gen.drop(df_gen[df_gen.isna().any(axis=1)].index)

            already_generated = df_gen.shape[0]

            # Update process bar
            progress.update(task, completed=min(already_generated, n_samples))

            tries += 1

        if tries == max_tries:
            raise RuntimeError(
                "max_tries reached, use a higher number, model did not converge or"
                " max_length is not high enough."
            )

        df_gen = df_gen.reset_index(drop=True)

        if self.categorical_columns == []:
            return df_gen.head(n_samples)
        else:
            df_inversed = self.decode_categorical_column(df_gen.head(n_samples))
            return df_inversed

    def tabula_sample(
        self,
        starting_prompts: tp.Union[str, list[str]],
        temperature: float = 0.7,
        max_length: int = 100,
        device: str = "cuda",
    ) -> pd.DataFrame:
        """Generate synthetic tabular data samples conditioned on a given input.

        Args:
            starting_prompts: String or List of Strings on which the output is conditioned.
             For example, "Sex is female, Age is 26"
            temperature: The generation samples each token from the probability distribution given by a softmax
             function. The temperature parameter controls the softmax function. A low temperature makes it sharper
             (0 equals greedy search), a high temperature brings more diversity but also uncertainty into the output.
             See this blog article (https://huggingface.co/blog/how-to-generate) to read more about the generation
             process.
            max_length: Maximal number of tokens to generate - has to be long enough to not cut any information
            device: Set to "cpu" if the GPU should not be used. You can also specify the concrete GPU.

         Returns:
            Pandas DataFrame with synthetic data generated based on starting_prompts
        """

        self.model.to(device)
        starting_prompts = (
            [starting_prompts]
            if isinstance(starting_prompts, str)
            else starting_prompts
        )
        generated_data = []

        # Generate a sample for each starting point
        for prompt in tqdm(starting_prompts):
            start_token = torch.tensor(self.tokenizer(prompt)["input_ids"]).to(device)

            # Generate tokens
            gen = self.model.generate(
                input_ids=torch.unsqueeze(start_token, 0),
                max_length=max_length,
                do_sample=True,
                temperature=temperature,
                pad_token_id=50256,
            )
            generated_data.append(torch.squeeze(gen))

        # Convert Text back to Tabular Data
        decoded_data = _convert_tokens_to_text(generated_data, self.tokenizer)
        df_gen = _convert_text_to_tabular_data(
            decoded_data, pd.DataFrame(columns=self.columns)
        )

        return df_gen

    def save(self, path: str):
        """Save Tabula Model

        Saves the model weights and a configuration file in the given directory.

        Args:
            path: Path where to save the model
        """
        # Make directory
        if os.path.isdir(path):
            warnings.warn(f"Directory {path} already exists and is overwritten now.")
        else:
            os.mkdir(path)

        # Save attributes
        with open(path + "/config.json", "w") as f:
            attributes = self.__dict__.copy()
            attributes.pop("tokenizer")
            attributes.pop("model")

            # NDArray is not JSON serializable and therefore has to be converted into a list.
            if isinstance(attributes["conditional_col_dist"], np.ndarray):
                attributes["conditional_col_dist"] = list(
                    attributes["conditional_col_dist"]
                )

            json.dump(attributes, f)

        # Save model weights
        torch.save(self.model.state_dict(), path + "/model.pt")

    def load_finetuned_model(self, path: str):
        """Load fine-tuned model

        Load the weights of a fine-tuned large language model into the Tabula pipeline

        Args:
            path: Path to the fine-tuned model
        """
        self.model.load_state_dict(torch.load(path))

    @classmethod
    def load_from_dir(cls, path: str):
        """Load Tabula class

        Load trained Tabula model from directory.

        Args:
            path: Directory where Tabula model is saved

        Returns:
            New instance of Tabula loaded from directory
        """
        assert os.path.isdir(path), f"Directory {path} does not exist."

        # Load attributes
        with open(path + "/config.json", "r") as f:
            attributes = json.load(f)

        # Create new be_tabula model instance
        tabula = cls(attributes["llm"])

        # Set all attributes
        for k, v in attributes.items():
            setattr(tabula, k, v)

        # Load model weights
        tabula.model.load_state_dict(torch.load(path + "/model.pt", map_location="cpu"))

        return tabula

    def _update_column_information(self, df: pd.DataFrame):
        # Update the column names (and numerical columns for some sanity checks after sampling)
        self.columns = df.columns.to_list()
        self.num_cols = df.select_dtypes(include=np.number).columns.to_list()

    def _update_conditional_information(
        self, df: pd.DataFrame, conditional_col: tp.Optional[str] = None
    ):
        assert conditional_col is None or isinstance(
            conditional_col, str
        ), f"The column name has to be a string and not {type(conditional_col)}"
        assert conditional_col is None or conditional_col in df.columns, (
            f"The column name {conditional_col} is not in the feature names of the"
            " given dataset"
        )

        # Take the distribution of the conditional column for a starting point in the generation process
        self.conditional_col = conditional_col if conditional_col else df.columns[-1]
        self.conditional_col_dist = _get_column_distribution(df, self.conditional_col)

    def _get_start_sampler(
        self,
        start_col: tp.Optional[str],
        start_col_dist: tp.Optional[tp.Union[tp.Dict, tp.List]],
    ) -> TabulaStart:
        if start_col and start_col_dist is None:
            raise ValueError(
                f"Start column {start_col} was given, but no corresponding"
                " distribution."
            )
        if start_col_dist is not None and not start_col:
            raise ValueError(
                f"Start column distribution {start_col} was given, the column name is"
                " missing."
            )

        assert start_col is None or isinstance(
            start_col, str
        ), f"The column name has to be a string and not {type(start_col)}"
        assert (
            start_col_dist is None
            or isinstance(start_col_dist, dict)
            or isinstance(start_col_dist, list)
        ), (
            "The distribution of the start column on has to be a list or a dict and"
            f" not {type(start_col_dist)}"
        )

        start_col = start_col if start_col else self.conditional_col
        start_col_dist = start_col_dist if start_col_dist else self.conditional_col_dist

        if isinstance(start_col_dist, dict):
            return CategoricalStart(self.tokenizer, start_col, start_col_dist)
        elif isinstance(start_col_dist, list):
            return ContinuousStart(self.tokenizer, start_col, start_col_dist)
        else:
            return RandomStart(self.tokenizer, self.columns)



from . import Tuner
from gentab.generators import TVAE
from gentab.evaluators import Evaluator
from gentab.utils import console, SPINNER, REFRESH

import optuna


class TVAETuner(Tuner):
    def __init__(
        self,
        evaluator: Evaluator,
        trials: int,
        *args,
        min_epochs: int = 300,
        max_epochs: int = 800,
        min_batch: int = 512,
        max_batch: int = 8192,
        max_tries_per_batch: int = 8192,
        timeout: int = None,
        **kwargs,
    ) -> None:
        super().__init__(
            evaluator,
            trials,
            min_epochs=min_epochs,
            max_epochs=max_epochs,
            min_batch=min_batch,
            max_batch=max_batch,
            max_tries_per_batch=max_tries_per_batch,
            timeout=timeout,
        )

    def objective(self, trial: optuna.trial.Trial) -> float:
        epochs = trial.suggest_int("epochs", self.min_epochs, self.max_epochs)
        batch_size = trial.suggest_categorical("batch_size", self.batch_sizes)
        compress_dims = (
            trial.suggest_categorical("comp_dims_in", [32, 64, 128, 256, 512]),
            trial.suggest_categorical("comp_dims_out", [32, 64, 128, 256, 512]),
        )
        decompress_dims = (
            trial.suggest_categorical("decomp_dims_in", [32, 64, 128, 256, 512]),
            trial.suggest_categorical("decomp_dims_out", [32, 64, 128, 256, 512]),
        )
        embedding_dim = trial.suggest_categorical("embedding_dim", [64, 128, 256, 512])
        l2scale = trial.suggest_float("l2scale", 1e-6, 1e-4, log=True)
        loss_factor = trial.suggest_categorical("loss_factor", [2, 4, 8, 16])
        pac = trial.suggest_categorical("pac", [2, 4, 8, 16, 32])

        self.generator = TVAE(
            self.dataset,
            epochs=epochs,
            batch_size=batch_size,
            embedding_dim=embedding_dim,
            compress_dims=compress_dims,
            decompress_dims=decompress_dims,
            l2scale=l2scale,
            loss_factor=loss_factor,
            pac=pac,
            max_tries_per_batch=self.max_tries_per_batch,
        )
        self.generator.generate()

        self.store_data(trial)

        acc, mcc = self.evaluator.evaluate(validation=True)

        return mcc



from . import Evaluator
from gentab.utils import console, SPINNER, REFRESH

import lightgbm as lgb


class LightGBM(Evaluator):
    def __init__(
        self,
        generator,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(generator, **kwargs)
        self.model = lgb.LGBMClassifier(
            *args,
            random_state=self.seed,
            verbosity=0,
            **kwargs,
        )
        self.callbacks = [lgb.log_evaluation(period=0)]

    def preprocess(self, X, y, X_test, y_test):
        X = self.dataset.encode_categories(X)
        X_test = self.dataset.encode_categories(X_test)

        return X, y, X_test, y_test
